#!/usr/bin/env python3
# encoding: utf-8


import ursgal
import glob
import os.path
import sys


def main(folder):
    '''

    Example script for reproducing the data for figure 3

    usage:

        ./barth_et_al_large_scale.py <folder>

    The folder determines the target folder where the files will be downloaded

    Chlamydomonas reinhardtii samples

    Three biological replicates of 4 conditions (2_3, 2_4, 3_1, 4_1)

    For more details on the samples please refer to
    Barth, J.; Bergner, S. V.; Jaeger, D.; Niehues, A.; Schulze, S.; Scholz,
    M.; Fufezan, C. The interplay of light and oxygen in the reactive oxygen
    stress response of Chlamydomonas reinhardtii dissected by quantitative mass
    spectrometry. MCP 2014, 13 (4), 969–989.

    Merge all search results (per biological replicate and condition, on folder
    level) on engine level and validate via percolator.

    'LTQ XL high res':

        * repetition 1
        * repetition 2

    'LTQ XL low res':

        * repetition 3

    Database:

        * Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta

    Note:

        The database and the files will be automatically downloaded from our
        webpage and peptideatlas
    '''

    input_params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'modifications': [
            'M,opt,any,Oxidation',
            '*,opt,Prot-N-term,Acetyl',  # N-Acetylation
        ],
        'ftp_url': 'ftp.peptideatlas.org',

        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',

        'ftp_output_folder_root': folder,
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        )
    }

    uc = ursgal.UController(
        params=input_params
    )

    if os.path.exists(input_params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    output_folder_to_file_list = {
        ('rep1_sample_2_3', 'LTQ XL high res'): [
            'CF_07062012_pH8_2_3A.mzML',
            'CF_13062012_pH3_2_3A.mzML',
            'CF_13062012_pH4_2_3A.mzML',
            'CF_13062012_pH5_2_3A.mzML',
            'CF_13062012_pH6_2_3A.mzML',
            'CF_13062012_pH11FT_2_3A.mzML',
        ],
        ('rep1_sample_2_4', 'LTQ XL high res'): [
            'CF_07062012_pH8_2_4A.mzML',
            'CF_13062012_pH3_2_4A_120615113039.mzML',
            'CF_13062012_pH4_2_4A.mzML',
            'CF_13062012_pH5_2_4A.mzML',
            'CF_13062012_pH6_2_4A.mzML',
            'CF_13062012_pH11FT_2_4A.mzML',

        ],
        ('rep1_sample_3_1', 'LTQ XL high res'): [
            'CF_12062012_pH8_1_3A.mzML',
            'CF_13062012_pH3_1_3A.mzML',
            'CF_13062012_pH4_1_3A.mzML',
            'CF_13062012_pH5_1_3A.mzML',
            'CF_13062012_pH6_1_3A.mzML',
            'CF_13062012_pH11FT_1_3A.mzML',
        ],
        ('rep1_sample_4_1', 'LTQ XL high res'): [
            'CF_07062012_pH8_1_4A.mzML',
            'CF_13062012_pH3_1_4A.mzML',
            'CF_13062012_pH4_1_4A.mzML',
            'CF_13062012_pH5_1_4A.mzML',
            'CF_13062012_pH6_1_4A.mzML',
            'CF_13062012_pH11FT_1_4A.mzML',
        ],

        ('rep2_sample_2_3', 'LTQ XL high res'): [
            'JB_18072012_2-3_A_FT.mzML',
            'JB_18072012_2-3_A_pH3.mzML',
            'JB_18072012_2-3_A_pH4.mzML',
            'JB_18072012_2-3_A_pH5.mzML',
            'JB_18072012_2-3_A_pH6.mzML',
            'JB_18072012_2-3_A_pH8.mzML',
        ],
        ('rep2_sample_2_4', 'LTQ XL high res'): [

            'JB_18072012_2-4_A_FT.mzML',
            'JB_18072012_2-4_A_pH3.mzML',
            'JB_18072012_2-4_A_pH4.mzML',
            'JB_18072012_2-4_A_pH5.mzML',
            'JB_18072012_2-4_A_pH6.mzML',
            'JB_18072012_2-4_A_pH8.mzML',

        ],
        ('rep2_sample_3_1', 'LTQ XL high res'): [
            'JB_18072012_3-1_A_FT.mzML',
            'JB_18072012_3-1_A_pH3.mzML',
            'JB_18072012_3-1_A_pH4.mzML',
            'JB_18072012_3-1_A_pH5.mzML',
            'JB_18072012_3-1_A_pH6.mzML',
            'JB_18072012_3-1_A_pH8.mzML',
        ],
        ('rep2_sample_4_1', 'LTQ XL high res'): [
            'JB_18072012_4-1_A_FT.mzML',
            'JB_18072012_4-1_A_pH3.mzML',
            'JB_18072012_4-1_A_pH4.mzML',
            'JB_18072012_4-1_A_pH5.mzML',
            'JB_18072012_4-1_A_pH6.mzML',
            'JB_18072012_4-1_A_pH8.mzML',
        ],

        ('rep3_sample_2_3', 'LTQ XL low res'): [
            'JB_FASP_pH3_2-3_28122012.mzML',
            'JB_FASP_pH4_2-3_28122012.mzML',
            'JB_FASP_pH5_2-3_28122012.mzML',
            'JB_FASP_pH6_2-3_28122012.mzML',
            'JB_FASP_pH8_2-3_28122012.mzML',
            'JB_FASP_pH11-FT_2-3_28122012.mzML',
        ],
        ('rep3_sample_2_4', 'LTQ XL low res'): [
            'JB_FASP_pH3_2-4_28122012.mzML',
            'JB_FASP_pH4_2-4_28122012.mzML',
            'JB_FASP_pH5_2-4_28122012.mzML',
            'JB_FASP_pH6_2-4_28122012.mzML',
            'JB_FASP_pH8_2-4_28122012.mzML',
            'JB_FASP_pH11-FT_2-4_28122012.mzML',
        ],
        ('rep3_sample_3_1', 'LTQ XL low res'): [
            'JB_FASP_pH3_3-1_28122012.mzML',
            'JB_FASP_pH4_3-1_28122012.mzML',
            'JB_FASP_pH5_3-1_28122012.mzML',
            'JB_FASP_pH6_3-1_28122012.mzML',
            'JB_FASP_pH8_3-1_28122012.mzML',
            'JB_FASP_pH11-FT_3-1_28122012.mzML',
        ],
        ('rep3_sample_4_1', 'LTQ XL low res'): [
            'JB_FASP_pH3_4-1_28122012.mzML',
            'JB_FASP_pH4_4-1_28122012.mzML',
            'JB_FASP_pH5_4-1_28122012.mzML',
            'JB_FASP_pH6_4-1_28122012.mzML',
            'JB_FASP_pH8_4-1_28122012.mzML',
            'JB_FASP_pH11-FT_4-1_28122012_130121201449.mzML',
        ],
    }

    for (outfolder, profile), mzML_file_list in sorted(output_folder_to_file_list.items()):
        uc.params['ftp_output_folder'] = os.path.join(
            input_params['ftp_output_folder_root'],
            outfolder
        )
        uc.params['ftp_include_ext'] = mzML_file_list

        if os.path.exists(uc.params['ftp_output_folder']) is False:
            os.makedirs(uc.params['ftp_output_folder'])

        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )

    if os.path.exists(input_params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    search_engines = [
        'omssa_2_1_9',
        'xtandem_piledriver',
        'myrimatch_2_1_138',
        'msgfplus_v9979',
        'msamanda_1_0_0_5243',
    ]

    # This dict will be populated with the percolator-validated results
    # of each engine ( 3 replicates x4 conditions = 12 files each )
    percolator_results = {
        'omssa_2_1_9': [],
        'xtandem_piledriver': [],
        'msgfplus_v9979': [],
        'myrimatch_2_1_138': [],
        'msamanda_1_0_0_5243': [],
    }

    five_files_for_venn_diagram = []

    for search_engine in search_engines:

        # This list will collect all 12 result files for each engine,
        # after Percolator validation and filtering for PSMs with a
        # FDR <= 0.01
        filtered_results_of_engine = []
        for mzML_dir_ext, mass_spectrometer in output_folder_to_file_list.keys():
            # for mass_spectrometer, replicate_dir in replicates:
            # for condition_dir in conditions:
            uc.set_profile(mass_spectrometer)

            mzML_dir = os.path.join(
                input_params['ftp_output_folder_root'],
                mzML_dir_ext
            )
            # i.e. /media/plan-f/mzML/Christian_Fufezan/ROS_Experiment_2012/Juni_2012/2_3/Tech_A/
            # all files ending with .mzml in that directory will be used!

            unified_results_list = []
            for filename in glob.glob(os.path.join(mzML_dir, '*.mzML')):
                # print(filename)
                if filename.lower().endswith(".mzml"):
                    # print(filename)
                    unified_search_results = uc.search(
                        input_file=filename,
                        engine=search_engine,
                    )
                    unified_results_list.append(
                        unified_search_results
                    )

            # Merging results from the 6 pH-fractions:
            merged_unified = uc.execute_misc_engine(
                input_file=unified_results_list,
                engine='merge_csvs_1_0_0',
            )

            # Validation with Percolator:
            percolator_validated = uc.validate(
                input_file=merged_unified,
                engine='percolator_2_08',  # one could replace this with 'qvality'
            )
            percolator_results[search_engine].append(
                percolator_validated
            )

            # At this point, the analysis is finished. We got
            # Percolator-validated results for each of the 3
            # replicates and 12 conditions.

            # But let's see how well the five search engines
            # performed! To compare, we collect all PSMs with
            # an estimated FDR <= 0.01 for each engine, and
            # plot this information with the VennDiagram UNode.
            # We will also use the Combine FDR Score method
            # to combine the results from all five engines,
            # and increase the number of identified peptides.

    five_large_merged = []
    filtered_final_results = []

    # We will estimate the FDR for all 60 files
    # (5 engines x12 files) when using percolator PEPs as
    # quality score
    uc.params['validation_score_field'] = 'PEP'
    uc.params['bigger_scores_better'] = False

    # To make obtain smaller CSV files (and make plotting
    # less RAM-intensive, we remove all decoys and PSMs above
    # 0.06 FDR
    uc.params['csv_filter_rules'] = [
        ['estimated_FDR', 'lte', 0.06],
        ['Is decoy', 'equals', 'false']
    ]
    for engine, percolator_validated_list in percolator_results.items():

        # unfiltered files for cFDR script
        twelve_merged = uc.execute_misc_engine(
            input_file=percolator_validated_list,
            engine='merge_csvs_1_0_0',
        )

        twelve_filtered = []
        for one_of_12 in percolator_validated_list:
            one_of_12_FDR = uc.validate(
                input_file=one_of_12,
                engine='add_estimated_fdr_1_0_0'
            )
            one_of_12_FDR_filtered = uc.execute_misc_engine(
                input_file=one_of_12_FDR,
                engine='filter_csv_1_0_0'
            )
            twelve_filtered.append(one_of_12_FDR_filtered)

        # For the combined FDR scoring, we merge all 12 files:
        filtered_merged = uc.execute_misc_engine(
            input_file=twelve_filtered,
            engine='merge_csvs_1_0_0'
        )

        five_large_merged.append(twelve_merged)
        filtered_final_results.append(filtered_merged)

    # The five big merged files of each engine are combined:
    cFDR = uc.combine_search_results(
        input_files=five_large_merged,
        engine='combine_FDR_0_1',
    )

    # We estimate the FDR of this combined approach:
    uc.params['validation_score_field'] = 'Combined FDR Score'
    uc.params['bigger_scores_better'] = False

    cFDR_FDR = uc.validate(
        input_file=cFDR,
        engine='add_estimated_fdr_1_0_0'
    )

    # Removing decoys and low quality hits, to obtain a
    # smaller file:
    uc.params['csv_filter_rules'] = [
        ['estimated_FDR', 'lte', 0.06],
        ['Is decoy', 'equals', 'false']
    ]
    cFDR_filtered_results = uc.execute_misc_engine(
        input_file=cFDR_FDR,
        engine='filter_csv_1_0_0',
    )
    filtered_final_results.append(cFDR_filtered_results)

    # Since we produced quite a lot of files, let's print the full
    # paths to our most important result files so we find them quickly:
    print(
        '''
These files can now be easily parsed and plotted with your
plotting tool of choice! We used the Python plotting library
matplotlib. Each unique combination of Sequence, modification
and charge was counted as a unique peptide.
        '''
    )
    print("\n########### Result files: ##############")
    for result_file in filtered_final_results:
        print(
            '\t*{0}'.format(
                result_file
            )
        )

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(main.__doc__)
        sys.exit(1)
    main(sys.argv[1])



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Example for plotting a simple Venn diagram with single ursgal csv files.

    usage:
        ./simple_venn_example.py


    '''
    uc = ursgal.UController(
        profile='LTQ XL low res',
        params={
            'visualization_label_positions': {
                '0': 'omssa',
                '1': 'xtandem'
            }

        }
    )

    file_list = [
        os.path.join(
            os.pardir,
            'tests',
            'data',
            'omssa_2_1_9',
            'test_BSA1_omssa_2_1_9.csv'
        ),
        os.path.join(
            os.pardir,
            'tests',
            'data',
            'xtandem_sledgehammer',
            'test_BSA1_xtandem_sledgehammer.csv'
        ),
    ]

    uc.visualize(
        input_files=file_list,
        engine='venndiagram_1_1_0',
        force=True,
    )
    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8
import ursgal
R = ursgal.UController()

TESTS = [
    {
        'input' : {
            'input_files' : [
                'uSearch_OSX',
                'uSearch_WINDOWS',
            ],
        },
        'output' : ('uSearch_' , '')
    },
    {
        'input' : {
            'input_files' : [
                'uSearch_OSX__ROCKS',
                'uSearch_WINDOWS__ROCKS',
            ],
        },
        'output' : ('uSearch_' , '__ROCKS')
    },
    {
        'input' : {
            'input_files' : [
                'uSearch__ROCKS',
                'not_only_uSearch__ROCKS',
            ],
        },
        'output' : ('' , 'uSearch__ROCKS')
    }

]


def determine_common_common_head_tail_part_of_names_test():
    for test_id, test_dict in enumerate(TESTS):
        yield determine_common_common_head_tail_part_of_names, test_dict


def determine_common_common_head_tail_part_of_names( test_dict ):
    out_put = R.determine_common_head_tail_part_of_names(
        **test_dict['input']
    )
    print('Results', out_put , test_dict)
    assert out_put == test_dict['output'], '''
    determine_common_name {0} failed'''.format(
        test_dict
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        determine_common_common_head_tail_part_of_names( test_dict )



#!/usr/bin/env python
'''
    usage:
        svm.py unified_input.csv engine_score_column_name
    i.e. :
        svm.py omssa_2_1_6_unified.csv 'OMSSA:pvalue'

    Writes a new file with added column "SVMscore" which is the distance to
    the separating hyperplane of a Percolator-like support vector machine.
'''
import numpy as np
import sklearn
from sklearn import svm
from sklearn.cross_validation import StratifiedKFold
from sklearn.preprocessing import Imputer
from collections import Counter, defaultdict
from random import random
import csv
import re
import os
import argparse

from misc import (get_score_colname_and_order, field_to_float,
                  unify_sequence,calc_FDR, scale_scores, row_is_decoy,
                  field_to_bayes_float, get_mz_values)






SCALER = sklearn.preprocessing.RobustScaler()  # RobustScaler() seems to be most robust ;)
PROTON = 1.00727646677


class SVMWrapper(dict):
    def __init__(self):
        self._svm_score_name = 'SVMscore'
        self.counter = {  # counting the # of possible training PSMs
            'target': 0,
            'decoy': 0,
            'positive': 0,
            'negative': 0,
            'unknown': 0,
            'parsed PSMs': 0,
        }
        self.results  = {}
        self.shitty_decoy_seqs = set()  # is overwritten by find_shitty_decoys()
        self.mgf_lookup = {}
        self.pep_to_mz = {}

        if __name__ == '__main__':
            self.parse_options()  # parse command line args and set options
            self.set_input_csv()

        self.observed_charges = set()
        self.used_extra_fields = set()
        self.decoy_train_prob = None  # probability to include decoy PSMs as negative training examples
        self.maximum_proteins_per_line = 0
        self.tryptic_aas = set(['R', 'K', '-'])
        self.delim_regex = re.compile(r'<\|>|\;')  # regex to split a line by both ";" and "<|>"
        return


    def parse_options(self):
        '''
        parses the command line args for options/parameters
        '''
        parser = argparse.ArgumentParser()
        parser.add_argument(
            '-i', '--input_csv', type=str, help='Input CSV path(s)',
            required=True, nargs='+')
        parser.add_argument(
            '-o', '--output_csv', type=str, help='Output CSV path',
            required=True)
        parser.add_argument(
            '-k', '--kernel', type=str, default='rbf',
            help='SVM kernel type ("rbf", "linear", "poly" or "sigmoid")')
        parser.add_argument(
            '-c', type=float, default=1.0,
            help='Penalty parameter C of the error term')
        parser.add_argument(
            '-g', '--gamma', type=str, default='auto',
            help='Gamma parameter of the SVM.')
        parser.add_argument(
            '-r', '--mb_ram', type=float, default=4000,
            help='Available RAM in megabytes, for SVM calculation')
        parser.add_argument(
            '-f', '--fdr_cutoff', type=float, default=0.01, 
            help='Target PSMs with a lower FDR will be used as a '\
                 'positive training set')
        parser.add_argument(
            '-x', '--columns_as_features', type=str, nargs='+', default=[
                'MS-GF:RawScore',
                'MS-GF:DeNovoScore',
                'MS-GF:SpecEValue',
                'MS-GF:EValue',
                'OMSSA:evalue',
                'OMSSA:pvalue',
                'X\!Tandem:expect',
                'X\!Tandem:hyperscore',
            ],
            help='Columns that should be used as a feature directly '\
                 '(e.g. secondary scores). Will be converted to float')
        parser.add_argument(
            '-d', '--dump_svm_matrix', type=str, default=False,
            help='Dump SVM matrix in PIN (Percolator input) format '
                 'to the specified path, mostly for debugging '
                 'and benchmarking.')

        arg_dict = vars(parser.parse_args())  # convert to dict
        self.update(arg_dict)
        try:
            self['gamma'] = float(self['gamma'])
        except ValueError:
            assert self['gamma'] == 'auto', 'Invalid gamma param: '\
                '"{0}", using "auto" instead.'.format(self['gamma'])


    def set_input_csv(self):
        '''
        distinguishes one vs. many unified input csv files and either
        sets the single csv as input, or merges all csvs and sets
        the merged csv as input.
        '''
        if len(self['input_csv']) > 1:
            raise Exception('You must only specify *one* unified CSV file!')
        self.csv_path = self['input_csv'][0]
        print('Using input file', self.csv_path)


    def find_shitty_decoys(self):
        '''
        Finds and notes decoys that share their sequence with a target PSM.
        
        Also counts the number of targets and decoys to get a quick estimate
        of how many positive/negative training examples can be "claimed".
        '''
        target_seqs = set()
        decoy_seqs = set()
        with open(self.csv_path, 'r') as f:

            reader = csv.DictReader(f)

            sorted_reader = sorted(
                reader, reverse=self['bigger_scores_better'],
                key=lambda d: float(d[self.col_for_sorting])
            )

            for row in sorted_reader:
                self.observed_charges.add(int(row['Charge']))
                if row_is_decoy(row):
                    decoy_seqs.add(unify_sequence(row['Sequence']))
                    self.counter['decoy'] += 1
                else:
                    target_seqs.add(unify_sequence(row['Sequence']))
                    self.counter['target'] += 1

        self.shitty_decoy_seqs = target_seqs.intersection(decoy_seqs)
        if len(self.shitty_decoy_seqs) > 0:
            print(
                'Warning! Found {0} sequences that are target AND decoy '\
                '(immutable peptides?). These will not be used for training.\n'.format(len(self.shitty_decoy_seqs))
            )
        return


    def determine_csv_sorting(self):
        with open(self.csv_path, 'r') as in_file:
            reader = csv.DictReader(in_file)
            self.col_for_sorting, self['bigger_scores_better'] = \
                get_score_colname_and_order(reader.fieldnames)
        if self.col_for_sorting == self._svm_score_name:
            self._svm_score_name = self._svm_score_name + '2'

        print('CSV will be sorted by column {0} (reverse={1}'\
              ')'.format(self.col_for_sorting, self['bigger_scores_better']))

        for feat in self['columns_as_features']:
            if feat in reader.fieldnames and feat != self.col_for_sorting:
                self.used_extra_fields.add(feat)


    def sort_by_rank(self, rowdict):
        score = float(rowdict[self.col_for_sorting])
        spec_title = rowdict['Spectrum Title']
        return (spec_title, score)


    @staticmethod
    def parse_protein_ids(csv_field, sep='<|>'):
        '''
        Turns the unified CSV column "Protein ID"
        into a set of all protein IDs.
        '''
        clean = csv_field.replace('decoy_', '').strip()
        prot_id_set = set(clean.split(sep))
        return prot_id_set


    def count_intra_set_features(self):
        '''
        intra-set features as calculated by Percolator:
        - num_pep:  Number of PSMs for which this is the best scoring peptide.
        - num_prot: Number of times the matched protein matches other PSMs.
        - pep_site: Number of different peptides that match this protein.

        own ideas:
        - pep_charge_states: in how many charge states was the peptide found?
        - seq_mods: in how many mod states was the AA-sequence found?
        - num_spec: Number of times the matched spectrum matches other peptides.
        '''
        print('Counting intra-set features...')
        self.num_pep = defaultdict(int)
        self.num_prot = defaultdict(set)
        self.pep_site = defaultdict(set)
        self.score_list_dict = defaultdict(list)

        self.pep_charge_states = defaultdict(set)
        self.seq_mods = defaultdict(set)
        self.num_spec = defaultdict(set)


        with open(self.csv_path, 'r') as f:
            reader = csv.DictReader(f)
            previous_spec_title = None
            rows_of_spectrum = []

            for row in sorted(
                reader,
                reverse=self['bigger_scores_better'],
                key=self.sort_by_rank
            ):

                if unify_sequence(row['Sequence']) in self.shitty_decoy_seqs:
                    continue
                current_spec_title = row['Spectrum Title']
                if current_spec_title != previous_spec_title:
                    # the next spectrum started, so let's process the info we
                    # collected for the previous spectrum:
                    score_list = [field_to_bayes_float(r[self.col_for_sorting]) for r in rows_of_spectrum]
                    self.score_list_dict[previous_spec_title] = score_list

                    for rank, line in enumerate(rows_of_spectrum):
                        #print("\t".join([
                            #str(rank), line['Spectrum Title'], line[self.col_for_sorting]
                        #]))
                        uni_sequence = unify_sequence(line['Sequence'])
                        peptide = (uni_sequence, line['Modifications'])

                        # multiple proteins are separated by <|>
                        # ignore start_stop_pre_post part since it depends on the peptide
                        # and not the protein (i.e. _233_243_A_R)
                        proteins = set(line['Protein ID'].replace('decoy_', '').split(';'))

                        #old unify csv format:
                        #proteins = self.parse_protein_ids(
                        #    line['proteinacc_start_stop_pre_post_;']
                        #)
                        if len(proteins) > self.maximum_proteins_per_line:
                            self.maximum_proteins_per_line = len(proteins)

                        if rank == 0:
                            # this is the 'best' peptide for that spectrum
                            self.num_pep[peptide] += 1
                        for protein in proteins:
                            self.num_prot[protein].add(
                                (line['Spectrum Title'], uni_sequence, line['Modifications'])
                            )
                            self.pep_site[protein].add(peptide)

                        self.pep_charge_states[peptide].add(int(row['Charge']))
                        self.seq_mods[uni_sequence].add(row['Modifications'])
                        self.num_spec[line['Spectrum Title']].add(peptide)

                    rows_of_spectrum = []

                rows_of_spectrum.append(row)
                previous_spec_title = current_spec_title


    def row_to_features(self, row):
        '''
        Converts a unified CSV row to a SVM feature matrix (numbers only!)
        '''
        sequence = unify_sequence(row['Sequence'])
        charge = field_to_float( row['Charge'] )
        score = field_to_bayes_float( row[self.col_for_sorting] )
        calc_mz, exp_mz, calc_mass, exp_mass = get_mz_values(row)
        #calc_mz = field_to_float( row['Calc m/z'] )  # calc m/z or uCalc?
        #exp_mz = field_to_float( row['Exp m/z'] )

        pre_aa_field = row['Sequence Pre AA']
        post_aa_field = row['Sequence Post AA']
        all_pre_aas = set(re.split(self.delim_regex, pre_aa_field))
        all_post_aas = set(re.split(self.delim_regex, post_aa_field))

        if any(pre_aa not in self.tryptic_aas for pre_aa in all_pre_aas):
            enzN = 0
        else:
            enzN = 1

        if any(post_aa not in self.tryptic_aas for post_aa in all_post_aas):
            enzC = 0
        else:
            enzC = 1

        n_missed_cleavages = len([aa for aa in sequence[:-1] if aa in ['R', 'K']])  # / len(sequence)

        missed_cleavages = [0] * 6
        try:
            missed_cleavages[n_missed_cleavages] = 1
        except IndexError:  # if a peptide has more than 6 missed cleavages
            missed_cleavages[-1] = 2

        spectrum = row['Spectrum Title'].strip()
        mass = (exp_mz * charge) - (charge - 1) * PROTON
        pep_len = len(sequence)
        #delta_mz = calc_mz - exp_mz
        delta_mass = calc_mass - exp_mass

        peptide = (sequence, row['Modifications'])
        proteins = self.parse_protein_ids(
            row['Protein ID']
        )
        num_pep = self.num_pep[peptide]
        pep_charge_states = len(self.pep_charge_states[peptide])
        seq_mods = len(self.seq_mods[sequence])
        num_spec = len(self.num_spec[row['Spectrum Title']])
        num_prot = sum(
            (len(self.num_prot[protein]) for protein in proteins)
        )
        pep_site = sum(
            (len(self.pep_site[protein]) for protein in proteins)
        )

        user_specified_features = []
        for feat in self.used_extra_fields:
            if feat != self.col_for_sorting:
                try:
                    user_specified_features.append(field_to_float(row[feat]))
                except ValueError:
                    pass

        charges = defaultdict(int)
        for charge_n in sorted(self.pep_charge_states[peptide]):
            charges[charge_n] = 1

        if sequence in self.shitty_decoy_seqs:
            is_shitty = 1
        else:
            is_shitty = 0

        score_list = sorted(
            list(set(self.score_list_dict[spectrum])),
            reverse=self['bigger_scores_better']
        )

        try:
            score_list_scaled = scale_scores(score_list)
            rank = score_list.index(score)
            deltLCn = score_list_scaled[rank] - score_list_scaled[ 1]  # Fractional difference between current and second best XCorr
            deltCn  = score_list_scaled[rank] - score_list_scaled[-1]  # Fractional difference between current and worst XCorr
        except (ValueError, IndexError, AssertionError):
            # NaN values will be replaced by the column mean later
            # NaN values are entered when there is no ranking
            # e.g. when only one peptide was matched to the spectrum.
            rank, deltLCn, deltCn = np.nan, np.nan, np.nan

        features = [
            score,
            rank,
            deltCn,
            deltLCn,
            charge,
            #delta_mz,# / pep_len,
            delta_mass,# / pep_len,
            #abs(delta_mz),# / pep_len,
            abs(delta_mass),# / pep_len,
            n_missed_cleavages / pep_len,
            missed_cleavages[0],
            missed_cleavages[1],
            missed_cleavages[2],
            missed_cleavages[3],
            missed_cleavages[4],
            missed_cleavages[5],
            enzN,
            enzC,
            mass,
            pep_len,
            num_pep,
            num_prot,
            pep_site,  
            is_shitty,
            pep_charge_states,
            num_spec,
            seq_mods,
        ]

        for charge_n in self.observed_charges:
            features.append(charges[charge_n])

        return features + user_specified_features


    def collect_data(self):
        '''
        parses a unified csv file and collects features from each row
        '''
        categories = []
        list_of_feature_lists = []
        feature_sets = set()
        with open(self.csv_path, 'r') as f:
            reader = csv.DictReader(f)
            # collecting some stats for FDR calculation:
            self.PSM_count   = 0
            self.decoy_count = 0

            if self['dump_svm_matrix']:
                self.init_svm_matrix_dump()
                additional_matrix_info = []

            for i, row in enumerate(
                sorted(reader, reverse=self['bigger_scores_better'],
                       key=lambda d: float(d[self.col_for_sorting])
                )):

                features = self.row_to_features(row)

                if tuple(features) in feature_sets:
                    continue
                feature_sets.add(tuple(features))

                category, psm_FDR = self.get_psm_category(row)

                list_of_feature_lists.append(features)
                categories.append(category)

                if self['dump_svm_matrix']:
                    label = -1 if row_is_decoy(row) else 1
                    sequence = '{0}.{1}#{2}.{3}'.format(
                        row['Sequence Pre AA'].strip(),
                        row['Sequence'].strip(),
                        row['Modifications'].strip(),
                        row['Sequence Post AA'].strip(),
                    )
                    additional_matrix_info.append({
                        'psm_id': row['Spectrum Title'].strip(),
                        'label': label,
                        'scannr': row['Spectrum Title'].strip().split('.')[-2],
                        'peptide': sequence,
                        'proteins': self.parse_protein_ids(row['Protein ID']),
                    })

                if i % 1000 == 0:
                    score_val = float(row[self.col_for_sorting])
                    msg = 'Generating feature matrix from input csv '\
                          '(line ~{0}) with score {1} and FDR '\
                          '{2}'.format(i, score_val, psm_FDR)
                    print(msg, end = '\r')


        # All data points are collected in one big matrix, to make standardization possible
        print('\nConverting feature matrix to NumPy array...')
        X_raw = np.array(list_of_feature_lists, dtype=float)

        print('Replacing empty/NaN values with the mean of each column...')
        self.nan_replacer = Imputer()
        self.nan_replacer.fit(X_raw)
        X_raw = self.nan_replacer.transform(X_raw)
        # Standardize input matrix to ease machine learning! Scaled data has zero mean and unit variance
        print('Standardizing input matrix...')
        self.scaler = SCALER.fit(X_raw)
        self.X = self.scaler.transform(X_raw)
        self.categories = np.array(categories)
        print()

        if self['dump_svm_matrix']:
            print('Dumping SVM matrix to', self['dump_svm_matrix'])
            
            for i, matrix_row in enumerate(self.X):
                matrix_row_info = additional_matrix_info[i]
                self.dump_svm_matrix_row(
                    row = list(matrix_row),
                    psm_id=matrix_row_info['psm_id'],
                    label=matrix_row_info['label'],
                    scannr=matrix_row_info['scannr'],
                    peptide=matrix_row_info['peptide'],
                    proteins=matrix_row_info['proteins'],
                )
            
            print('Dumped SVM matrix to', self['dump_svm_matrix'])
        return


    def init_svm_matrix_dump(self):
        from misc import FEATURE_NAMES
        colnames = ['PSMId', 'label', 'scannr'] + FEATURE_NAMES
        colnames += ["charge{0}".format(c) for c in self.observed_charges]
        for extra_field in sorted(self.used_extra_fields):
            colnames += [extra_field]
        colnames += ['peptide']
        for n_proteins in range(self.maximum_proteins_per_line):
            colnames.append('proteinId{0}'.format(n_proteins + 1))
        self.matrix_csv_path = self['dump_svm_matrix']
        print('Dumping raw SVM input matrix to', self.matrix_csv_path)
        with open(self.matrix_csv_path, 'w') as f:
            f.write('\t'.join(colnames) + '\n')
        

    def dump_svm_matrix_row(self, row=None, psm_id=None, label=None, scannr=None, peptide=None, proteins=None):
        full_row = [psm_id, label, scannr] + row + [peptide] + list(proteins)
        with open(self.matrix_csv_path, 'a') as f:
            row_str = '\t'.join(str(x) for x in full_row) + '\n'
            f.write(row_str)


    def get_psm_category(self, row):
        '''
        Determines whether a PSM (csv row) should be used as a negative or
        positive training example.

        returns
            1  - high-scoring target (positive training example)
            0  - not-high-scoring target (not usable for training)
           -1  - decoy (negative training example)
        
        '''
        category = 0  # unknown (mix of true positives and false positives)
        self.PSM_count += 1  # for FDR calculation
        sequence = unify_sequence(row['Sequence'])
        psm_FDR = calc_FDR(self.PSM_count, self.decoy_count)

        if row_is_decoy(row):
            self.decoy_count += 1
            if psm_FDR <= 0.25 and sequence not in self.shitty_decoy_seqs:
                category = -1  # decoy (false positive hits)
                self.counter['negative'] += 1
            else:
                if not self.decoy_train_prob:
                    need_max = self.counter['positive'] * 2
                    have = self.counter['negative']
                    still_there = self.counter['decoy'] - have
                    prob = need_max / still_there
                    if prob < 0.001:
                        prob = 0.001
                    self.decoy_train_prob = prob
                    print()
                    print(self.counter)
                    print('need max:', need_max)
                    print('have:', have)
                    print('still_there:', still_there)
                    print('probability:', self.decoy_train_prob)
                    print()
                if self.decoy_train_prob >= 1.0 or random() <= self.decoy_train_prob:
                    category = -1  # decoy (false positive hits)
                    self.counter['negative'] += 1

                

        else:  # row is target
            if psm_FDR <= self['fdr_cutoff'] and sequence not in self.shitty_decoy_seqs:
                category = 1  # high quality target (almost certainly true positives)
                self.counter['positive'] += 1

        if category == 0:
            self.counter['unknown'] += 1
        return (category, psm_FDR)


    def train(self, training_matrix, training_categories):
        counter = Counter(training_categories)
        msg = 'Training {0} SVM on {1} target PSMs and {2} decoy PSMs'\
        '...'.format(self['kernel'], counter[1], counter[-1])
        print(msg, end = '\r')
        # specify the classification method (rbf and linear SVC seem to work best and are quite fast)
        classifier = svm.SVC(
            C           = self['c'],
            kernel      = self['kernel'],
            probability = False,  # we don't want to get probabilities later on -> faster
            cache_size  = self['mb_ram'],  # available RAM in megabytes
            #decision_function_shape = 'ovr',  # doesn't seem to matter
            #class_weight= 'balanced',  # doesn't seem to matter
        )
        # train the SVC on our set of training data:
        classifier.fit(
            training_matrix,
            training_categories,
        )
        print(msg + ' done!')
        return classifier


    def classify(self, classifier, psm_matrix):
        msg = 'Classifying {0} PSMs...'.format(len(psm_matrix))
        print(msg, end = '\r')
        for i, row in enumerate(psm_matrix):
            # get the distance to the separating SVM hyperplane and use it as a score:
            svm_score = classifier.decision_function(np.array([row]))[0]

            features = tuple(row)
            if features not in self.results:
                self.results[features] = svm_score
            else:
                print(
                    'Warning! This combination of features already has a predicted probability! '\
                    'Previous svm_score: {0:f} - Current svm_score: {1:f}'\
                    ''.format( self.results[tuple(row)], svm_score )
                )
                # take the mean value, no idea how to handle this better, but it never happened so far...
                self.results[features] = (self.results[features] + svm_score) / 2.0
        print(msg + ' done!')
        return


    def add_scores_to_csv(self):
        outfname = os.path.basename(self['output_csv'])
        print('Writing output csv {0} ...'.format(outfname))
        msg = 'Writing output csv {0} (line ~{1})...'

        with open(self['output_csv'], 'w', newline='') as out_csv, open(self.csv_path, 'r') as in_csv:
            reader = csv.DictReader(in_csv)
            writer = csv.DictWriter(out_csv, reader.fieldnames + [self._svm_score_name])
            writer.writeheader()
            for i, row in enumerate(reader):
                if i % 1000 == 0:
                    print(msg.format(outfname, i), end='\r')
                features = self.nan_replacer.transform(
                    np.array([ self.row_to_features(row) ])
                )
                features_scaled = tuple(list(self.scaler.transform(features)[0]))
                SVMScore = self.results[features_scaled]
                row[self._svm_score_name] = SVMScore
                writer.writerow(row)
        print('\n')
        return


    def __str__(self):
        out_str = ['\n\tpyPercolator Options:']
        for option, value in self.items():
            out_str.append(
                '{0:·<25}{1}'.format(option, value)
            )
        return '\n'.join(out_str)



if __name__ == '__main__':
    s = SVMWrapper()

    print(s)  # print parameter/settings overview

    s.determine_csv_sorting()
    s.find_shitty_decoys()
    
    print('\nCounter:')
    print(s.counter)
    print()
    s.count_intra_set_features()
    s.collect_data()

    print('Splitting data in half to avoid training and testing on the same features...')
    skfold = StratifiedKFold(
        s.categories, n_folds = 2, shuffle = True
    )

    # use one half to score the other half, and vice versa:
    for i, (train_index, test_index) in enumerate(skfold):
        current_half = '1st' if i == 0 else '2nd'
        other_half = '2nd' if i == 0 else '1st'
        print('\nUsing high-scoring PSMs and decoys of the {0} half to train...'.format(current_half))
        mask = s.categories[train_index] != 0
        train_categories = s.categories[train_index][mask]
        train_features = s.X[train_index][mask]
        svm_classifier = s.train(
            training_matrix     = train_features,
            training_categories = train_categories,
        )


        print('Using the trained SVM to classify all PSMs of the {0} half'.format(other_half))
        s.classify(
            svm_classifier,
            s.X[test_index],
        )
        if s['kernel'].lower() == 'linear':
            print()  # print SVM coefficients (only works for linear kernel)
            print(svm_classifier.coef_)
            print()

    print('\nCounter:')
    print(s.counter)
    print()
    s.add_scores_to_csv()



#!/usr/bin/env python
'''
Converts mzML to mgf

Usage:
./_mzml_2_mgf.py <mzML File1> ... <mzML FileN>

Note:
1) A new mgf file will be created at the same location as the mzML file
2) only ms2 are converted to mgf
'''
# 18.11.2014
# By C. Fufezan

from __future__ import print_function
import sys
import os
import argparse
import pymzml

try:
    version = pymzml.__version__
except AttributeError:
    version = '0.x.x'
if version.split('.')[0] != '2':
    raise RuntimeError(
        'Using mzml2mgf converter V2, however pymzML version {version} is installed\n'.format(version=version),
        'Please install pymzML version 0.x.x or set the mzml2mgf_converter_version param to mzml2mgf_1_0_0'
    )


def _determine_mzml_name_base(file_name, prefix):
    file_name = os.path.basename(file_name)
    if file_name.upper().endswith('.MZML.GZ'):
        mzml_name_base = file_name[:-8]
    elif file_name.upper().endswith('.MZML'):
        mzml_name_base = file_name[:-5]
    elif file_name.upper().endswith('.IDX.GZ'):
        mzml_name_base = file_name[:-7]
    else:
        raise Exception(
            "Can not determine mzml base name from {0}".format(
                file_name
            )
        )
    if prefix is not None and prefix is not '':
        mzml_name_base = '_'.join([prefix, mzml_name_base])
    return mzml_name_base


def main(
    mzml=None,
    mgf=None,
    i_decimals=5,
    mz_decimals=5,
    machine_offset_in_ppm=None,
    scan_exclusion_list=None,
    scan_inclusion_list=None,
    prefix=None,
    scan_skip_modulo_step=None,
    ms_level=2,
    precursor_min_charge=1,
    precursor_max_charge=5,
    ion_mode='+',
    spec_id_attribute=None,
):

    print(
        'Converting file:\n\tmzml : {0}\n\tto\n\tmgf : {1}'.format(
            mzml,
            mgf,
        )
    )
    mzml_name_base = _determine_mzml_name_base(mzml, prefix)
    # rt_lookup = mzml_name_base + '_rt_lookup.pkl'
    oof = open(mgf , 'w')
    run = pymzml.run.Reader(
        mzml
    )
    tmp = {
        'rt_2_scan' : {},
        'scan_2_rt' : {},
        'scan_2_mz': {}
    }
    mgf_entries = 0
    if scan_exclusion_list is None:
        scan_exclusion_list = []
    else:
        scan_exclusion_list = [int(spec_id) for spec_id in scan_exclusion_list]

    if machine_offset_in_ppm is not None:
        mz_correction_factor = machine_offset_in_ppm * 1e-6
    else:
        mz_correction_factor = 0
    precursor_charge_range = '{0}'.format(precursor_min_charge)
    for charge in range(precursor_min_charge + 1, precursor_max_charge + 1):
        precursor_charge_range += ' and {0}'.format(charge)
    mzml_basename = os.path.basename(mzml)
    for n, spec in enumerate(run):
        if n % 500 == 0:
            print(
                'File : {0:^40} : Processing spectrum {1}'.format(
                    mzml_basename,
                    n,
                ),
                end='\r'
            )

        spec_ms_level = spec.ms_level
        scan_time, scan_time_unit = spec.scan_time
        if scan_time_unit == 'seconds':
            scan_time /= 60
        elif scan_time_unit != 'minute':
            print('''
                [Warning] The retention time unit is not recognized or not specified.
                [Warning] It is assumed to be minutes and continues with that.
            ''')
        if spec_id_attribute is None:
            spectrum_id = spec.ID
        else:
            id_attribute, id_key = list(spec_id_attribute.items())[0]
            if id_attribute == 'ID':
                spectrum_id = spec.ID
            elif id_attribute == 'index':
                spectrum_id = spec.index+1
            elif id_attribute == 'id_dict':
                spectrum_id = spec.id_dict[id_key]
            else:
                print('''
                    [ERROR] Please specifiy an available spec_id_attribute for mzml2mgf.
                    [ERROR] Available: ID, id_dict, index
                ''')
        tmp['rt_2_scan'][scan_time] = spectrum_id
        tmp['scan_2_rt'][spectrum_id] = scan_time
        tmp['unit'] = 'minute'

        if spec_ms_level != ms_level:
            continue

        peaks_2_write = spec.peaks('centroided')

        precursor_mz = spec.selected_precursors[0]['mz']
        precursor_charge = spec.selected_precursors[0].get('charge', None)

        if scan_inclusion_list is not None:
            if int(spectrum_id) not in scan_inclusion_list:
                continue

        if int(spectrum_id) in scan_exclusion_list:
            continue
        mgf_entries += 1

        if scan_skip_modulo_step is not None:
            if mgf_entries % scan_skip_modulo_step != 0:
                continue

        print(
            'BEGIN IONS',
            file=oof
        )
        print(
            'TITLE={0}.{1}.{1}.{2}'.format(
                mzml_name_base,
                spectrum_id,
                precursor_charge,
            ),
            file=oof
        )
        print(
            'SCANS={0}'.format(
                spectrum_id
            ),
            file=oof
        )

        scan_time = float(scan_time) * 60
        print(
            'RTINSECONDS={0}'.format(
                round(
                    scan_time,
                    11
                )
            ),
            file=oof
        )

        precursor_mz += precursor_mz * mz_correction_factor
        tmp['scan_2_mz'][spectrum_id] = precursor_mz
        print(
            'PEPMASS={0}'.format(
                precursor_mz
            ),
            file=oof
        )
        if precursor_charge is not None:
            print(
                'CHARGE={0}{1}'.format(
                    precursor_charge,
                    ion_mode
                ),
                file=oof
            )
        else:
            print(
                'CHARGE={0}{1}'.format(
                    precursor_charge_range,
                    ion_mode
                ),
                file=oof
            )

        for mz, intensity in peaks_2_write:
            # if fragment_ppm_offset is not None:
            mz += mz * mz_correction_factor
            print(
                '{0:<10.{mzDecimals}f} {1:<10.{intensityDecimals}f}'.format(
                    mz,
                    intensity,
                    mzDecimals=mz_decimals,
                    intensityDecimals=i_decimals
                ),
                file=oof
            )

        print(
            'END IONS\n',
            file=oof
        )
    print('')
    print(
        'Wrote {0} mgf entries to file {1}'.format(
            mgf_entries,
            mgf
        )
    )

    oof.close()
    return tmp


if __name__ == '__main__':
    # parsing command line arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'mzml',
        help='mzml input file')
    parser.add_argument(
        'mgf',
        help='mzml output file')
    parser.add_argument(
        '-m', '--mz_decimals', default=5,
        help='Number of decimals for mz values', type=int)
    parser.add_argument(
        '-i', '--i_decimals', default=5,
        help='Number of decimals for intensity values', type=int)

    if len(sys.argv) <= 1:
        parser.print_help()
    else:
        args = parser.parse_args()
        tmp = main(**args.__dict__)
        # print(tmp.keys())
        # print(sorted( tmp['scan_2_rt'].keys() ))



#!/usr/bin/env python

import ahocorasick


FASTA = [
    ('>MoA', 'A' * 5),
    ('>MoB', 'B' * 5),
    ('>MoC', 'C' * 5)
]

ALL_SEQS = '>>'.join( [ seq for (_id, seq) in FASTA ] )
# our mean delimiter that spares us the overlapping check :))
seqidx_to_fastaidx = {}

fullseq_pos = 0
for fasta_index, (_id, seq) in enumerate(FASTA):
    for x in range(len(seq)):
        fullseq_pos += 1
        seqidx_to_fastaidx[ fullseq_pos ] = fasta_index
    fullseq_pos += 2
    # our mean delimiter that spares us the overlapping check :))
# even sweeter with index bisecting in list of starting positions

au = ahocorasick.Automaton()
au.add_word('BBBB', ['BBBB', ':)'])
au.add_word('AABB', ['AABB', ':)'])
au.add_word('CCCC', ['CCCC', ':)'])
au.make_automaton()

for match in au.iter( ALL_SEQS ):
    seq_index, not_sure_what_do_with_this = match
    fastidx = seqidx_to_fastaidx[ seq_index ]
    print(match, 'match on', FASTA[ fastidx ])



#!/usr/bin/env python
import ursgal
import os


class mgf_to_rt_lookup_1_0_0(ursgal.UNode):
    """Write RT lookup from mgf file
    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'mgf_to_rt_lookup',
        'version'                     : '1.0.0',
        'engine_type': {
            'converter': True
        },
        'output_extensions'           : ['.txt'],
        'create_own_folder'           : False,
        'utranslation_style'          : 'mgf_to_rt_lookup_style_1',
        'citation'                    : '',
        'in_development'              : False,
        'include_in_git'              : True,
        'input_extensions'            : ['.mgf'],
        'release_date'                : None,
        'engine': {
            'platform_independent': {
                'arc_independent': {
                    'exe': 'mgf_to_rt_lookup_1_0_0.py',
                },
            },
        },
        'citation': '',
    }

    def __init__(self, *args, **kwargs):
        super(mgf_to_rt_lookup_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        print('[ -ENGINE- ] Executing conversion ..')
        self.time_point(tag='execution')
        main = self.import_engine_as_python_function()
        if self.params['input_file'].lower().endswith('.mgf') is False:
            raise ValueError('Trying to convert a non-mgf file')

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        input_file = os.path.join(
                self.params['input_dir_path'],
                self.params['input_file']
            )

        main(
            input_file=input_file,
            output_file=output_file,
            rt_pickle_name=self.params['rt_pickle_name'],
            spec_id_regex=r'\w+.(\d+).\d+.\d',
            run_id_regex=r'(\w+).\d+.\d+.\d'
        )
        # TODO add spec_id_regex and run_id_regex as uparams, could also be used in unify_csv etc
        self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python3.2
# encoding: utf-8
'''
Small test to verify the correct function of the isotope mass calculation
in chemical_composition
'''
import ursgal

TESTS = [
    {
        'composition_1' : {
            '14N' : 1,
            '12C' : 1
        },
        'composition_2' : {
            'N' : 1,
            'C' : 1
        },
    },
    {
        'composition_1' : {
            'N'   : 1,
            '14N' : 1,
            '12C' : 1
        },
        'composition_2' : {
            'N' : 2,
            'C' : 1
        },
    },
]

cc_1 = ursgal.ChemicalComposition()
cc_2 = ursgal.ChemicalComposition()


def pepitde_with_unimod_test():
    for test_id, test_dict in enumerate( TESTS ):
        yield mass_checker, test_dict


def mass_checker( test_dict ):

    cc_1.add_chemical_formula(
        test_dict['composition_1']
    )
    cc_2.add_chemical_formula(
        test_dict['composition_2']
    )
    assert cc_1._mass() == cc_2._mass()
    cc_1.clear()
    cc_2.clear()


if __name__ == '__main__':
    print(__doc__)
    pepitde_with_unimod_test()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import glob
import sys


def main(folder=None, profile=None, target_decoy_database=None):
    '''
    An example test script to search all mzML files which are present in the
    specified folder and post-process them with a support vector machine.
    This method is similar to Percolator. Furthermore, SVM scores are used
    to estimate the false discovery rate (FDR) of PSMs. This allows the user
    to choose an error rate.

    Note:
        SVM post-processing requires the free Python module scikit-learn.
        For installation instructions, see http://scikit-learn.org/stable/install.html

    The machine profile has to be specified as well as the target-decoy
    database.

    usage:

        ./simple_svm_post_processing_folder_wide.py <mzML_folder> <profile> <target_decoy_database>


    Current profiles:

        * 'QExactive+'
        * 'LTQ XL low res'
        * 'LTQ XL high res'

    '''
    input_params = {
        'database': target_decoy_database,
        'modifications': [
            'M,opt,any,Oxidation',
            '*,opt,Prot-N-term,Acetyl',  # N-Acetylation
        ],

        # You can customize the SVM kernel and parameters like this:

        'fdr_cutoff': 0.01,  # target PSMs with a lower FDR than this
        # threshold will be used as a positive training set in the SVM
        'kernel': 'rbf',  # the kernel function of the SVM ("rbf",
        # "linear", "poly" or "sigmoid")
        'svm_c_param': 1.0,  # penalty parameter C of the error
        # term of the SVM
    }

    uc = ursgal.UController(
        profile=profile,
        params=input_params,
    )

    engine_list = [
        'omssa',
        'xtandem_vengeance',
        'msgfplus_v2016_09_16',
        'myrimatch_2_1_138',
    ]

    # find all .mzML files in the folder that was
    # specified as first command line argument
    mzML_files = glob.glob(os.path.join(folder, '*.mzML'))

    svm_validated_files = []

    for search_engine in engine_list:
        unified_csvs = []
        for mzML_file in mzML_files:
            unified_search_result_file = uc.search(
                input_file=mzML_file,
                engine=search_engine,
            )
            unified_csvs.append(unified_search_result_file)

        merged_unified_csv = uc.execute_misc_engine(
            input_file=unified_csvs,
            engine='merge_csvs'
        )

        # Statistical post-processing of PSMs with a support vector machine
        svm_validated_csv = uc.validate(
            input_file=merged_unified_csv,
            engine='svm_1_0_0',
        )

        # Estimate the false discovery rate (FDR) for each PSM,
        # based on the support vector machine's score
        uc.params['validation_score_field'] = 'SVMscore'
        # Larger SVM scores indicate higher confidence
        uc.params['bigger_scores_better'] = True
        svm_results_with_fdr = uc.add_estimated_fdr(
            input_file=svm_validated_csv,
        )
        # Reset validation parameters to default (for subsequent SVM
        # validation)
        uc.params['validation_score_field'] = None
        uc.params['bigger_scores_better'] = None
        svm_validated_files.append(svm_results_with_fdr)

    print('\nYour SVM-validated search results:')
    for svm_validated_file in svm_validated_files:
        print(svm_validated_file)
    return


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print(main.__doc__)
        sys.exit('You did not specify three command line arguments, exiting.')
    main(
        folder=sys.argv[1],
        profile=sys.argv[2],
        target_decoy_database=sys.argv[3]
    )



#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import pprint
import sys
import random
import time
import re


AAs = 'ACDEFGHIKLMNPQRSTVWY'


def generate_random_peptides(min_size=6, max_size=11, n=100000):
    peptides = []
    start = time.time()
    for iteration in range(n):
        aa_len = random.randrange(min_size, max_size)
        peptide = ''
        for _ in range(aa_len):
            peptide += random.choice(AAs)
        peptides.append(peptide)
    stop = time.time()
    print('\nPeptide generation took {0:.3f}s ..'.format(stop - start))
    return peptides


def main(fasta_file, mapper_class_version):
    '''
    Maps 100.000 peptides using regex or upeptide_master

    Usage:
    ./map_peptides_to_fasta_comparison.py <fasta_file> <class_name>

    Class name can be
        * 'UPeptideMapper_v4'
        * 'UPeptideMapper_v3'
        * 'UPeptideMapper_v2'

    '''
    print()
    ITERATIONS = 100000
    peptides = generate_random_peptides(n=ITERATIONS)
    print(
        'Generate {0}  peptides'.format(
            len(peptides)
        )
    )
    print(
        'Generate {0} unique peptides'.format(
            len(set(peptides))
        )
    )
    uc = ursgal.UController(verbose=False)

    start = time.time()
    upapa_class = uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function(
        mapper_class_version
    )
    if mapper_class_version == 'UPeptideMapper_v3':
        upeptide_mapper = upapa_class(fasta_file)
        fasta_name = upeptide_mapper.fasta_name
        args = [
            list(peptides),
            fasta_name
        ]
    elif mapper_class_version == 'UPeptideMapper_v4':
        upeptide_mapper = upapa_class(fasta_file)
        args = [
            list(peptides)
        ]
    else:
        upeptide_mapper = upapa_class(word_len=6)
        fasta_name = upeptide_mapper.build_lookup_from_file(fasta_file)
        args = [
            list(peptides),
            fasta_name
        ]

    stop = time.time()
    print(
        'With class {0} fasta Lookup generation took {1:.3f}s'.format(
            mapper_class_version,
            stop - start
        )
    )

    # exit(1)
    start = time.time()
    # for peptide in peptides:
    upeptide_mapper.map_peptides(*args)
    stop = time.time()
    print(
        'With class {0} peptide mapping took {1:.3f}s'.format(
            mapper_class_version,
            stop - start
        )
    )


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(main.__doc__)
    else:
        main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python
# encoding: utf-8
"""
    Ursgal
    ----------

    :copyright: (c) 2014 by C. Fufezan, J. Leufken, S. Schulze
    :licence: BSD, see LISCENSE for more details

"""

from __future__ import absolute_import
import sys
import re
from collections import defaultdict as ddict
import copy
import ursgal


class ChemicalComposition(dict):
    '''
    Chemical composition class. The actual sequence or formula can be reset
    using the `add` function.

    Keyword Arguments:
        sequence (str): Peptide or chemical formula sequence
        aa_compositions (Optional[dict]): amino acid compositions
        isotopic_distributions (Optional[dict]): isotopic distributions

    Keyword argument examples:

        **sequence** - Currently this can for example be::
              [
              '+H2O2H2-OH',
              '+{0}'.format('H2O'),
              '{peptide}'.format(pepitde='ELVISLIVES'),
              '{peptide}+{0}'.format('PO3', peptide='ELVISLIVES'),
              '{peptide}#{unimod}:{pos}'.format(
              peptide = 'ELVISLIVES',
              unimod = 'Oxidation',
              pos = 1
              )
              ]

    Examples::
        >>> c = ursgal.ChemicalComposition()
        >>> c.use("ELVISLIVES#Acetyl:1")
        >>> c.hill_notation()
        'C52H90N10O18'
        >>> c.hill_notation_unimod()
        'C(52)H(90)N(10)O(18)'
        >>> c
        {'O': 18, 'H': 90, 'C': 52, 'N': 10}
        >>> c.composition_of_mod_at_pos[1]
        defaultdict(<class 'int'>, {'O': 1, 'H': 2, 'C': 2})
        >>> c.composition_of_aa_at_pos[1]
        {'O': 3, 'H': 7, 'C': 5, 'N': 1}
        >>> c.composition_at_pos[1]
        defaultdict(<class 'int'>, {'O': 4, 'H': 9, 'C': 7, 'N': 1})

        >>> c = ursgal.ChemicalComposition('+H2O2H2')
        >>> c
        {'O': 2, 'H': 4}
        >>> c.subtract_chemical_formula('H3')
        >>> c
        {'O': 2, 'H': 1}

    Note:
        We did not include mass calculation, since pyQms will calculate masses
        much more accurately using unimod and other element enrichments.

    '''

    def __init__(
        self,
        sequence=None,
        aa_compositions=None,
        isotopic_distributions=None,
        monosaccharide_compositions=None,
    ):

        self._unimod_parser = None
        self.composition_of_mod_at_pos = {}
        """dict: chemical composition of unimod modifications at given position
        (if peptide sequence was used as input or using the `use` function)

        Note:
            Numbering starts at position 1, since all PSM search engines
            use this nomenclature.
        """
        self.composition_of_aa_at_pos = {}
        """dict: chemical composition of amino acid at given peptide position
        (if peptide sequence was used as input or using the `use` function)

        Note:
            Numbering starts at position 1, since all PSM search engines
            use this nomenclature.
        """
#                 Examples::
#                     c.composition_of_mod_at_pos[1] = {'15N': 2, '13C': 6, 'N': -2, 'C': -6}

#         """
        self.composition_at_pos = {}
        """dict: chemical composition at given peptide position incl modifications
        (if peptide sequence was used as input or using the `use` function)

        Note:
            Numbering starts at position 1, since all PSM search engines
            use this nomenclature.
        """
        self.peptide = None
        self.addon = None
        self.unimod_at_pos = {}
        # self.regex_patterns = {
        #     ':pos' : re.compile( r''':(?P<pos>[0-9]*)''' ),
        #     'aaN'  : re.compile( r'''(?P<aa>[A-Z]{1})(?P<N>[0-9]*)''' ),
        # }
        if aa_compositions is None:
            self.aa_compositions = ursgal.chemical_composition_kb.aa_compositions
        else:
            self.aa_compositions = aa_compositions
        if isotopic_distributions is None:
            self.isotopic_distributions = \
                ursgal.chemical_composition_kb.isotopic_distributions
        else:
            self.isotopic_distributions = isotopic_distributions
        if monosaccharide_compositions is None:
            self.monosaccharide_compositions = ursgal.chemical_composition_kb.monosaccharide_compositions
        else:
            self.monosaccharide_compositions = monosaccharide_compositions
        if sequence is not None:
            self.use(sequence)

        self.isotope_mass_lookup = {}
        for element, isotope_list in self.isotopic_distributions.items():
            for isotope_mass, abundance in isotope_list:
                isotope_mass_key = '{0}{1}'.format(
                    str(round(isotope_mass)).split('.')[0],
                    element
                )
                self.isotope_mass_lookup[isotope_mass_key] = isotope_mass

    def __add__(self, other_cc):
        '''Experimental'''
        tmp = copy.deepcopy(self)
        for other_key, other_value in other_cc.items():
            tmp[other_key] += other_value
        return tmp

    def __missing__(self, key):
        if key not in self.keys():
            self[key] = 0
        return self[key]

    def clear(self):
        '''Resets all lookup dictionaries and self

        One class instance can be used analysing a series of sequences, thereby
        avoiding class instantiation overhead
        '''

        self.composition_of_mod_at_pos.clear()
        self.composition_of_aa_at_pos.clear()
        self.composition_at_pos.clear()
        self.unimod_at_pos.clear()

        self.peptide = None
        self.addon = None
        for k in list(self.keys()):
            del self[k]

    def _parse_sequence_old_style(self, sequence):
        positions = [len(sequence)]
        for sign in ['+', '-']:
            if sign in sequence:
                positions.append(sequence.index(sign))
        minPos = min(positions)
        peptide = sequence[:minPos]
        addon = sequence[minPos:]
        self.peptide = peptide
        self.addon = addon
        if peptide != '':
            self.add_peptide(peptide)
            self['O'] += 1
            self['H'] += 2

        chemical_formula_blocks = re.compile(r'''
                        [+|-]{1}
                        [^-+]*
                        ''', re.VERBOSE).findall(addon)
        for cb in chemical_formula_blocks:
            if cb[0] == '+':
                self.add_chemical_formula(cb[1:])
            else:
                self.subtract_chemical_formula(cb[1:])
        return

    def _parse_sequence_unimod_style(self, sequence):
        minPos = sequence.index("#")
        peptide = sequence[:minPos]
        addon = sequence[minPos + 1:]
        self.peptide = peptide
        if peptide != '':
            self.add_peptide(peptide)
            self['O'] += 1
            self['H'] += 2
        self.addon = addon
        unimods = self.addon.split(';')
        # pattern = self.regex_patterns[':pos']
        pattern = re.compile( r''':(?P<pos>[0-9]*$)''' )
        for unimod in unimods:
            if unimod == '':
                continue
            unimod = unimod.strip()
            if ':' not in unimod:
                sys.exit(
                    'This unimod: {0} requires positional information'.format(unimod))

            for occ, match in enumerate(pattern.finditer(unimod)):
                try:
                    unimodcomposition = self._unimod_parser.name2composition(
                        unimod[:match.start()]
                    )
                except:
                    sys.exit(
                        'Cannot map unimod {0}. extracted position argument {1}'.format(
                            unimod, match.start()
                        ))
                # if occ >= 1:
                position = int(match.group('pos'))
                if position in self.unimod_at_pos.keys():
                    print('{0} <<- Two unimods at the same position ? '.format(
                        sequence
                    ))
                    raise Exception

                self.unimod_at_pos[position] = unimod[:match.start()]
            # match = re.search( position_re_pattern, unimod)
            # if match is not None:
            #     end = match.start()
            #     print( '>>>>', match)
            # else:
            #     end = len( unimod )
            # try:
            #     unimodcomposition = self._unimod_parser.name2composition(
            #         unimod[:end ]
            #     )
            # except:
            #     print(
            #         'Unimod error:', unimod,'>>', unimod[:end],
            #         re.search( position_re_pattern , unimod),
            #         re.search( position_re_pattern , unimod).start()
            #     )
            #     exit(1)
            # print( self , 'peptide only')
            # print( 'Unimod:', unimod, unimod[:end] , )
            # Full addition
            # print( unimodcomposition , '<<<<<<')
            for k, v in unimodcomposition.items():
                self[k] += v
            # storage position related modifications
            position = int(match.group('pos'))
            if position == 0:
                # E.g. Acetylation at pos 0 indicates N-Term
                # but has to be counted for position 1 in this class
                position = 1

            if position not in self.composition_of_mod_at_pos.keys():
                self.composition_of_mod_at_pos[position] = ddict(int)
            if position not in self.composition_at_pos.keys():
                self.composition_at_pos[position] = ddict(int)
            for k, v in unimodcomposition.items():
                self.composition_of_mod_at_pos[position][k] += v
                self.composition_at_pos[position][k] += v

        return

    def use(self, sequence):
        '''Re-initialize the class with a new sequence

        This is helpful if one wants to use the same class instance
        for multiple sequence since it remove class instantiation overhead.

        Args:
            sequence (str): See top for possible input formats.
        '''

        self.clear()
        # reset the shiznit
        if '#' in sequence:
            # Unimod Style format
            if self._unimod_parser is None:
                self._unimod_parser = ursgal.UnimodMapper()
            self._parse_sequence_unimod_style(sequence)
        else:
            self._parse_sequence_old_style(sequence)

    def add_chemical_formula(self, chemical_formula, factor=1):
        '''Adds chemical formula to the instance

        Args:
            chemical_formula (str): chemical composition given as Hill notation

        Keyword Arguments:
            factor (int): multiplication factor to add the same chemical formula
                multiple times
        '''
        self._merge(chemical_formula, mode='addition', factor=factor)
        return

    def add_peptide(self, peptide):
        """Adds peptide sequence to the instance"""
        # pattern = self.regex_patterns['aaN']
        pattern = re.compile( r'''(?P<aa>[A-Z]{1})(?P<N>[0-9]*)''' )
        # pattern = re.compile(r'[A-Z]{1}[0-9]*')
        number_offset = 0
        # this are the count for e.g. SILAC aa, i.e. R0 R1 or C0 and so on ...
        # print( peptide, type( peptide ))
        for aaN_match in pattern.finditer(peptide):
            aa = aaN_match.group('aa')
            N = aaN_match.group('N')
            pos = int(aaN_match.start()) - number_offset + 1
            if N != '':
                number_offset += len(N)
            try:
                aa_compo = self.aa_compositions[aa + N]
            except:
                sys.exit(
                    '''
                    Do not know aa composition for {0}
                    in {1}
                    '''.format(aa + N, peptide))
            self.add_chemical_formula(aa_compo)

            composition = self._chemical_formula_to_dict(aa_compo)
            self.composition_of_aa_at_pos[pos] = composition
            if pos not in self.composition_at_pos.keys():
                self.composition_at_pos[pos] = ddict(int)
            for k, v in composition.items():
                self.composition_at_pos[pos][k] += v

    def add_glycan(self, glycan):
        '''Adds a glycan to the instance.

        Args:
            glycan (str): sequence of monosaccharides given in unimod format,
                e.g.: HexNAc(2)Hex(3)dHex(1)Pent(1),
                available monosaccharides are listed in chemical_composition_kb
        '''
        pattern = re.compile(
            r'''(?P<monosacch>[A-z]*)(?P<count>\([0-9]*\))''' )
        for glyc_match in pattern.finditer(glycan):
            monosacch = glyc_match.group('monosacch')
            if glyc_match.group('count') == '':
                count = 1
            else:
                count = int(glyc_match.group('count').strip('(').strip(')'))
            # try: 
            #     monosacch_compo = self._unimod_parser.name2composition(monosacch)
            # except:
            if monosacch in self.monosaccharide_compositions.keys():
                monosacch_compo = self.monosaccharide_compositions[monosacch]
            else:
                sys.exit('Do not know glycan composition for {0}'.format(monosacch))
            self.add_chemical_formula(monosacch_compo, factor=count)

    def _chemical_formula_to_dict(self, chemical_formula):
        if isinstance(chemical_formula, ursgal.ChemicalComposition):
            chem_dict = chemical_formula
        elif '(' in chemical_formula:
            pattern = re.compile(
                r'(?P<element>[0-9]*[A-z]*)(?P<count>[(][0-9]*[)])')
        else:
            pattern = re.compile(r'(?P<element>[A-Z][a-z]*)(?P<count>[0-9]*)')
        chem_dict = {}
        # print( chemical_formula , type( chemical_formula ))
        for match in pattern.finditer(chemical_formula):
            if match.group('count') == '':
                count = 1
            else:
                count = int(match.group('count').strip('(').strip(')'))
            if match.group('element') not in chem_dict.keys():
                chem_dict[match.group('element')] = 0
            chem_dict[match.group('element')] += count
        return chem_dict

    def hill_notation(self, include_ones=False, cc=None):
        '''
        Formats chemical composition into `Hill notation`_ string.

        .. _Hill Notation:
            https://en.wikipedia.org/wiki/Hill_system

        Args:
            cc (dict, optional): can format other element dicts as well.

        Returns:
            str: Hill notation format of self.
                For examples::
                    C50H88N10O17
        '''
        MAJORS = ['C', 'H']
        s = ''
        if cc is None:
            cc_dict = self
        else:
            cc_dict = cc

        for major in MAJORS:
            if major in cc_dict.keys():
                if cc_dict[major] == 0:
                    continue
                s += major
                if include_ones or cc_dict[major] > 1:
                    s += str(cc_dict[major])
        for k in sorted(cc_dict.keys()):
            if k not in MAJORS:
                if cc_dict[k] == 0:
                    continue
                s += k
                if include_ones or cc_dict[k] > 1:
                    s += str(cc_dict[k])
        return s

    # def hill_notation(self, include_ones=False):
    def hill_notation_unimod(self, cc=None):
        '''
        Formats chemical composition into `Hill notation`_ string
        adding `unimod`_ features.

        .. _Hill Notation:
            https://en.wikipedia.org/wiki/Hill_system

        .. _unimod:
            http://www.unimod.org/fields.html

        Args:
            cc (dict, optional): can format other element dicts as well.

        Returns:
            str: Hill notation format including unimod format rules of self.
                For example::
                    C(50)H(88)N(10)O(17)

        '''
        MAJORS = ['C', 'H']
        s = ''
        if cc is None:
            cc_dict = self
        else:
            cc_dict = cc

        for major in MAJORS:
            if major in cc_dict.keys():
                if cc_dict[major] == 0:
                    continue
                s += '{0}({1})'.format(
                    major.replace('(', '').replace(')', ''),
                    cc_dict[major]
                )
        for k in sorted(cc_dict.keys()):
            if k not in MAJORS:
                if cc_dict[k] == 0:
                    continue
                s += '{0}({1})'.format(
                    k.replace('(', '').replace(')', ''),
                    cc_dict[k]
                )
        return s

    def _mass(self, cc=None):
        '''
        Calculate the mass of the chemical composition.
        Optional cc can be specified, i.e. a cc dict in the style of
        { element : count , ... }

        This does however not work with enriched elements, e.g. 15N
        Better use pyQms isotopologue library for more accurate mass
        calculations.
        '''
        mass = 0
        if cc is None:
            cc_mass_dict = self
        else:
            cc_mass_dict = cc
        for element, count in cc_mass_dict.items():
            isotope_mass = None
            try:
                isotope_mass = self.isotopic_distributions[element][0][0]
            except:
                # we check for 15N or 13C or other isotopes
                assert element in self.isotope_mass_lookup.keys(), '''
                The given element {0}
                is not included yet. Please add it to the isotopic_distributions
                dictionary in ursgal/chemical_composition_kb.py.
                Isotopic masses and distributions can be found at
                http://ciaaw.org/atomic-masses.html
                '''.format( element )
                isotope_mass = self.isotope_mass_lookup[element]
            mass += count * isotope_mass

        return mass

    def _merge(self, chemical_formula, mode='addition', factor=1):
        '''Generalized function that allows addition and subtraction'''
        if mode == 'addition':
            sign = +1
        elif mode == 'subtraction':
            sign = -1
        else:
            sys.exit('Do not know which mode to use for _merge')
        if isinstance(chemical_formula, str):
            chemical_formula = self._chemical_formula_to_dict(chemical_formula)
        for element, count in chemical_formula.items():
            self[element] = self[element] + sign * count * factor
        # else:
        #     print(chemical_formula, type(chemical_formula))
        #     sys.exit('Do not know the format of the chemical formula')
        return

    def subtract_peptide(self, peptide):
        '''Subtract peptide from instance'''
        for aa in peptide:
            self.subtract_chemical_formula(self.aa_compositions[aa])

    def subtract_chemical_formula(self, chemical_formula, factor=1):
        '''Subtract chemical formula from instance

        Args:
            chemical_formula (str): chemical composition given as Hill notation

        Keyword Arguments:
            factor (int): multiplication factor to add the same chemical formula
                multiple times
        '''
        self._merge(chemical_formula, mode='subtraction', factor=factor)
        return

    def generate_cc_dict(self):
        tmp = {}
        tmp.update(self)
        return tmp


if __name__ == '__main__':
    print('Yes!')



#!/usr/bin/env python

import ursgal
import os.path


class combine_pep_1_0_0(ursgal.UNode):
    '''
    combine_pep_1_0_0 UNode

    Combining Multiengine Search Results with "Combined PEP"

    "Combined PEP" is a hybrid approach combining elements of the
    "combined FDR" approach (Jones et al., 2009), elements of PeptideShaker,
    and elements of Bayes' theorem. Similar to "combined FDR", "combined PEP"
    groups the PSMs. For each search engine, the reported PSMs are treated
    as a set and the logical combinations of all sets are treated separately
    as done in the "combined FDR" approach. For instance, three search engines
    would result in seven PSM groups, which can be visualized by the seven
    intersections of a three-set Venn diagram. Typically, a PSM group that is
    shared by multiple engines contains fewer decoy hits and thus represents a
    higher quality subset and thus its PSMs receive a higher score. This
    approach is based on the assumption that the search engines agree on the
    decoys and false-positives as they agree on the targets.

    The combined PEP approach uses Bayes' theorem to calculate a multiengine
    PEP (MEP) for each PSM based on the PEPs reported by, for example,
    Percolator for different search engines, that is

    .. image:: http://pubs.acs.org/appl/literatum/publisher/achs/journals/content/jprobs/2016/jprobs.2016.15.issue-3/acs.jproteome.5b00860/20160229/images/pr-2015-00860d_m001.gif
       :target: http://pubs.acs.org/doi/full/10.1021/acs.jproteome.5b00860#_i2
       :align: center

    This is done for each PSM group separately.

    Then, the combined PEP (the final score) is computed similar to PeptideShaker
    using a sliding window over all PSMs within each group (sorted by MEP). Each
    PSM receives a PEP based on the target/decoy ratio of the surrounding PSMs.

    .. image:: http://pubs.acs.org/appl/literatum/publisher/achs/journals/content/jprobs/2016/jprobs.2016.15.issue-3/acs.jproteome.5b00860/20160229/images/pr-2015-00860d_m002.gif
       :target: http://pubs.acs.org/doi/full/10.1021/acs.jproteome.5b00860#_i2
       :align: center

    Finally, all groups are merged and the results reported in one output,
    including all the search result scores from the individual search engines
    as well as the FDR based on the "combined PEP".

    The sliding window size can be defined by adjusting the Ursgal parameter
    "window_size" (default is 249).

    Input should be multiple CSV files from different search engines. Each
    CSV requires a PEP column, for instance by post-processing with Percolator.

    Returns a merged CSV file with all PSMs that were found and two added
    columns:

    - column "Bayes PEP":
        The multi-engine PEP, see explanation above
    - column "combined PEP":
        The PEP as computed within the engine combination PSMs

    For optimal ranking, PSMs should be sorted by combined PEP.
    Ties can be resolved by sorting them by Bayes PEP.
    '''

    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'Combine PEP',
        'version'            : '1.0.0',
        'release_date'       : '2009-5-1',
        'engine_type' : {
            'meta_engine' : True,
        },
        'input_extensions'   : ['.csv'],
        'output_extensions'  : ['.csv'],
        'in_development'     : False,
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'combine_pep_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'combine_pep_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Combines PEP scores from different search engines.',
    }

    def __init__(self, *args, **kwargs):
        super(combine_pep_1_0_0, self).__init__(*args, **kwargs)
        pass


    def preflight( self ):
        '''
        Building the list of parameters that will be passed to the
        combine_pep_1_0_0 main function.

        These parameters are stored in self.command_dict

        Returns:
                None
        '''
        input_file_list_for_cPEP = []
        search_engine_list_for_cPEP = []
        # careful, these must have the same order!

        for input_file_dict in self.params["input_file_dicts"]:
            input_file_path = os.path.join(
                input_file_dict['dir'],
                input_file_dict['file']
            )
            input_file_list_for_cPEP.append(input_file_path)
            search_engine = input_file_dict["last_engine"]
            search_engine_list_for_cPEP.append(search_engine)

        features_that_define_unique_psm = \
            ['Sequence', 'Modifications', 'Spectrum Title', 'Is decoy']#, 'Charge', 'Is decoy']

        output_csv_full_path = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file'],
        )

        self.command_dict = {
            'input_csvs': input_file_list_for_cPEP,
            'input_engines': search_engine_list_for_cPEP,
            'output_csv': output_csv_full_path,
            'columns_for_grouping': features_that_define_unique_psm,
            'input_sep': ',',  # input csv separating char
            'output_sep': ',',  # output csv separating char
            'join_sep': ';',  # char to join multiple values in same field
            'pep_colname': 'PEP',
            # The size of the sliding window for PEP calculation:
            'window_size': self.params['translations']['window_size'],
        }

    def _execute( self ):
        '''
        Executing the combine_FDR_0_1 main function with parameters
        that were defined in preflight (stored in self.command_dict)

        The main function is imported and then executed using the
        parameters from command_dict.

        Returns:
                None
        '''
        cPEP_main = self.import_engine_as_python_function()

        print('''

Executing main() function from {scriptpath} with the following parameters:

>>> {script}.main(
...     input_csvs = {input_csvs},
...     input_engines = {input_engines},
...     output_csv = '{output_csv}',
...     columns_for_grouping = {columns_for_grouping},
...     input_sep = '{input_sep}',
...     output_sep = '{output_sep}',
...     join_sep = '{join_sep}',
...     pep_colname = '{pep_colname}',
...     window_size = {window_size},
... )

        '''.format(
            scriptpath = os.path.relpath(self.exe),
            script = os.path.basename(self.exe),
            **self.command_dict
            )
        )

        # executing main function of the combine_PEP script
        cPEP_main(
            input_csvs           = self.command_dict['input_csvs'],
            input_engines        = self.command_dict['input_engines'],
            output_csv           = self.command_dict['output_csv'],
            columns_for_grouping = self.command_dict['columns_for_grouping'],
            input_sep            = self.command_dict['input_sep'],
            output_sep           = self.command_dict['output_sep'],
            join_sep             = self.command_dict['join_sep'],
            pep_colname          = self.command_dict['pep_colname'],
            window_size          = self.command_dict['window_size'],
        )



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import csv

# TESTS =[
#     {'peptide': 'STGGAAGMLGGARSQRVVR'},
#     {'peptide': 'FTMADGGSIEPQENTSGEAFVEELIR'},
#     {'peptide': 'AAGGAIPHHTYAAACIDTEDVEAAAAACR'},
#     {'peptide': 'AERTGGEWGVSRDGAGGGSAGGGGR'},
#     {'peptide': 'QDAAVTTGTVAAAAAAAAFGGGK'},
#     {'peptide': 'AMVAGGVEAGVLADGADERSAVQLGK'},
#     {'peptide': 'IEGEFAAVEELDSSAGLLK'},
#     {'peptide': 'EGMRPRFTSSVPGEYR'},
#     {'peptide': 'AKAGASAADGTAASTNSESAEGRNSAEEK'},
#     {'peptide': 'ATSTESDEIALLEK'},
#     {'peptide': 'SPATGPATSGGLMSASPGNGAMVR'},
#     {'peptide': 'VEEDGEELSLDEESDALAASDFSQAEPR'},
#     {'peptide': 'RRGVGGGSGGSCGGGSGGQGGCGGGGAGR'},
#     {'peptide': 'AAAAAAPPADETAAAAAALEAVATAAAAPAPEAAPVGR'},
#     {'peptide': 'APGAGPGTTLLITDIQSSTALWELLPEDVMAK'},
#     {'peptide': 'AVSRANGGGGGGAAAAANTGGAAASTSGGGAAAEGAGAAGEGGEDPAMR'},
#     {'peptide': 'SSGGGGSRGAIGTGTGGANKFPQPK'},
#     {'peptide': 'GLLLDLADPVIGAAWEPHISDPGLWRFLKPHLK'},
#     {'peptide': 'GWLAVQALEPQAERREAPAPGPTAAVSAAAPAAADK'},
#     {'peptide': 'FQAAPWTFFLRSPLTDEPVDASMCR'},
#     {'peptide': 'VFQRVTVNAVDEWDEEARMHVAER'},
#     {'peptide': 'LGALEDADQGFSLLGIGFDEWK'},
#     {'peptide': 'CMNPVTDIAGGAEALLPSPTAPPQGSHSEVAR'},
#     {'peptide': 'GGSGGGGGGGGGGDGSGGRMVPGSGGSSAQAGQGPGDSLAGAPSASAAGR'},
#     {'peptide': 'QTQLEDQAEEEEAEAQAQAEK'},
#     {'peptide': 'KAFLEYLAVQQEQLAREGVAELPPEQQEQLMR'},
#     {'peptide': 'QLKDIEERSSAVSSGITENEEADLIEIIVISLNK'},
#     {'peptide': 'LGSTSAPPSASKHGTPAGGGGSGGGR'},
#     {'peptide': 'LGLFFPAGLAHTEPDRSKEPQYAQCGTR'},
#     {'peptide': 'VHGAYERYPTFLVQAHYEAGRGAAAEFMGYSGR'},
#     {'peptide': 'DLNEIPAAVQIILALELFSAVQDPMHLRDAKVK'},
#     {'peptide': 'GSSCSGTYVSGSALDVTHTGTSGQDSSGDVWGSAAEKRSGGR'},
#     {'peptide': 'LDFGMPGYLDGAKVAVEMRAGGLQLR'},
#     {'peptide': 'SEVAQPVGVEGSPSALLPAGLGSLVAGK'},
#     {'peptide': 'AAAAATQAGSRDPAHSLGKAAAAASLAAALLDASLLTTLHVHVPSGR'},
# ]

# R = ursgal.UController(
#     profile = 'LTQ XL low res',
#     params = {
#         'database': 'tests/data/test_Creinhardtii_target_decoy.fasta',
#         'modifications':[
#             '*,opt,Prot-N-term,Acetyl'    # N-Acteylation[]
#         ]
#     },
#     force=True
# )

# mgf_file = R.convert_to_mgf_and_update_rt_lookup(
#     input_file = 'tests/data/test_Creinhardtii_QE_pH8.mzML',
#     force      = True
# )
# raw_results = R.search_mgf(
#     input_file = mgf_file,
#     engine     = 'omssa',
#     force      = True
# )
# converted_results = R.convert_results_to_csv(
#     input_file = raw_results,
#     force      = True,
# )
# output_file = R.unify_csv(
#     input_file = converted_results,
#     force      = True
# )

# peptides_set = set()
# for line_dict in csv.DictReader(open(output_file, 'r')):
#     peptides_set.add( line_dict['Sequence'] )


# def compare_peptide_test():
#     for test_id, test_dict in enumerate(TESTS):
#         yield compare_peptide, test_dict


# def compare_peptide( test_dict ):
#     assert test_dict['peptide'] in peptides_set

# def number_of_peptides_test():
#     test_peptides = set()
#     for peptide_dict in TESTS:
#         test_peptides.add( peptide_dict['peptide'] )
#     assert len(peptides_set) == len(test_peptides)


# if __name__ == '__main__':
#     print(__doc__)



#!/usr/bin/env python
'''
collection of misc. functions and data that is imported into svm.py
'''
import numpy as np
import functools
import operator
import hashlib
import os


COLUMNS_TO_BE_REMOVED_FROM_MERGED_CSV = set([
    'Start',
    'Stop',
    'gi',
    'NIST score',
    'PSM_ID',
    'Pass Threshold',
    'Raw data location',
    'number of matched peaks',
    'number of unmatched peaks',
    'rank',
    'Rank',
    'Accession',
    'Filename',
])


FEATURE_NAMES = [
    "score",
    "rank",
    "deltCn",
    "deltLCn",
    "charge",
    "delta_mass",
    "abs_delta_mass",
    "n_missed_cleavages",
    "missed_cleavages[0]",
    "missed_cleavages[1]",
    "missed_cleavages[2]",
    "missed_cleavages[3]",
    "missed_cleavages[4]",
    "missed_cleavages[5]",
    "enzN",
    "enzC",
    "mass",
    "pep_len",
    "num_pep",
    "num_prot",
    "pep_site",  
    "is_shitty",
    "pep_charge_states",
    "num_spec",
    "seq_mods",
    #"i_ratio",
    #"a_ratio",
    #"b_ratio",
    #"y_ratio",
    #"MH_ratio",
    #"OMSSA:evalue",
]


def naive_bayes(prob_list):
    '''
    Combines independent probabilities a, b, c
    like this:

                            a*b*c
    combined_prob = -------------------------
                    a*b*c + (1-a)*(1-b)*(1-c)

    For a straightforward explanation, see
    http://www.paulgraham.com/naivebayes.html
    '''
    # multiplying all probabilities: a*b*c
    multiplied_probs = \
        functools.reduce(operator.mul, prob_list, 1)
    # multiplying the opposite of all probabilities: (1-a)*(1-b)*(1-c)
    multiplied_opposite_probs = \
        functools.reduce(operator.mul, (1-p for p in prob_list), 1)
    return multiplied_probs / (multiplied_probs + multiplied_opposite_probs)


def field_to_float(field):
    '''
    Converts the value of a field to float.
    If the field contains multiple separated floats,
    the mean is returned instead. If the field is empty,
    numpy.NaN is returned.
    '''
    try:
        result = float(field)
    except ValueError:
        if field == '':
            result = np.NaN
        elif ';' in field:
            values = [float(v) for v in field.split(';')]
            result = sum(values) / len(values)
        else:
            raise Exception(
                'Field value {0} cannot be converted to float.'.format(field)
            )
    return float(result)


def field_to_bayes_float(field):
    '''
    Converts the value of a field to float.
    If the field contains multiple separated floats,
    the naive Bayes combined value is returned instead.
    If the field is empty, numpy.NaN is returned.
    '''
    try:
        result = float(field)
    except ValueError:
        if field == '':
            result = np.NaN
        elif ';' in field:
            values = [float(v) for v in field.split(';')]
            result = naive_bayes(values)
        else:
            raise Exception(
                'Field value {0} cannot be converted to float.'.format(field)
            )
    return float(result)


def get_score_colname_and_order(colnames):
    '''
    Searches CSV headers for known score columns.
    '''
    known_scores = [
        # score_colname, bigger_scores_better
        ('estimated_FDR', False),
        ('SVMProb', False),
        ('OMSSA:pvalue', False),
        ('X\!Tandem:expect', False),
        ('MS-GF:SpecEValue', False),
        ('MyriMatch:MVH', True),
        ('Amanda:Score', True),
    ]
    for score, bsb in known_scores:
        if score in colnames:
            score_colname = score
            bigger_scores_better = bsb
            break
    else:
        raise Exception('Could not find a valid score field '\
            'for FDR estimation in file', input_csv)
    return score_colname, bigger_scores_better


def calc_FDR(PSM_count, false_positives):
    '''
    calculate false discovery rate according to FDR Method 2
    (Käll et al. 2007) as explained by Jones et al. 2009
    '''
    true_positives  = PSM_count - (2 * false_positives)
    if true_positives <= 0:  # prevent FDR above 1. Not sure if this is done?
        return 1.0
    FDR = false_positives / (false_positives + true_positives)
    return FDR


def scale_scores(l):
    # scales a list l to values from -1 to 1
    out_l = []
    assert len(l) > 1
    norm = max(l) - min(l)
    assert norm != 0
    for x in l:
        x_str = (x - min(l)) / norm
        out_l.append( (x_str * 2) -1 )
    return out_l


def unify_sequence(seq):
    '''
    Some sequences cannot be distinguished by MS (i.e. sequences
    where L is replaced by I and vice versa).
    Such target/decoy pairs are not suitable for training.
    This function applies rules to a sequence that allow identification 
    of such target/decoy pairs,
    i.e.
        unify_sequence('EILE') == unify_sequence('ELLE')
        -> True
    '''
    # replace leucin with isoleucin cause they cant be distinguished:
    seq = seq.replace('L', 'I')
    # sort first 2 amino acids:
    if len(seq) >= 2:
        seq = ''.join(sorted(seq[:2])) + seq[2:]
    return seq


def row_is_decoy(row):
    if row['Is decoy'].lower() == 'true':
        is_decoy = True
    elif row['Is decoy'].lower() == 'false':
        is_decoy = False
    else:
        raise Exception('Could not determine whether PSM is decoy or not.')
    if 'decoy_' in row.get('proteinacc_start_stop_pre_post_;', ''):
        is_decoy = True
    return is_decoy


def get_mz_values(row):
    calc_mz = field_to_float(row['Calc m/z'])
    exp_mz = field_to_float(row['Exp m/z'])
    ucalc_mz = field_to_float(row['uCalc m/z'])
    charge = field_to_float(row['Charge'])
    if int(round(exp_mz/ucalc_mz, 0)) == charge:
        # OMSSA returns masses instead of m/z values for some reason
        # dividing by charge to correct this
        calc_mass = calc_mz
        exp_mass = exp_mz
        calc_mz /= charge
        exp_mz /= charge
    else:  # no correction required, just computing the masses
        calc_mass = calc_mz * charge
        exp_mass = exp_mz * charge
    return calc_mz, exp_mz, calc_mass, exp_mass


def count_ion_partners(ion_d, pep_len, print_it=False):
    score = 0
    if print_it:
        print('\n ion partners:')
    for i in range(pep_len-2):
        y_ion = 'y' + str(i+1)
        a_ion = 'a' + str(pep_len-i-1)
        b_ion = 'b' + str(pep_len-i-1)

        ions = (y_ion, a_ion, b_ion)
        row_score = sum( int(bool(ion_d[i])) for i in ions ) ** 2
        score += row_score

        if print_it:
            print(int(bool(ion_d[y_ion])), int(bool(ion_d[a_ion])), int(bool(ion_d[b_ion])), ' - ', y_ion, a_ion, b_ion, row_score)
    return score


def hash_input_filenames(fname_list):
    fnames = ''.join([os.path.basename(f) for f in fname_list]).encode()
    hash_object = hashlib.md5(fnames)
    return hash_object.hexdigest()


if __name__ == '__main__':
    print('only for importing :)')



#!/usr/bin/env python
import ursgal
import os


class msgfplus2csv_py_v1_0_0(ursgal.UNode):
    """
    msgfplus2csv_py v1.0.0 UNode
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'msgfplus_py_mzid2csv',
        'version'            : 'v1.0.0',
        'release_date'       : '2018-08-28',
        'engine_type' : {
            'converter'     : True
        },
        'input_extensions'   : ['.mzid', '.mzid.gz'],
        'output_extensions'  : ['.csv'],
        'output_suffix'      : None,
        'include_in_git'     : True,
        'in_development'     : False,
        'distributable'      : True,
        'utranslation_style' : 'msgfplus_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe'     : 'msgfplus2csv_py_v1_0_0.py'
                },
            },
        },
        'citation' : ''
    }

    def __init__(self, *args, **kwargs):
        super(msgfplus2csv_py_v1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        msgf_mzid2csv = self.import_engine_as_python_function()

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )

        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        msgf_mzid2csv(input_file, output_file)
        return output_file



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Examples script for filtering validated results for a PEP <= 0.01 and
    remove all decoys.

    usage:
        ./filter_csv_validation_example.py


    Will produce a file with only target sequences with a posterior error
    probability of lower or equal to 1 percent
    '''
    params = {
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ]
    }

    csv_file_to_filter = os.path.join(
        os.pardir,
        'example_data',
        'misc',
        'filter_csv_for_mods_example_omssa_2_1_9_pmap_unified_percolator_2_08_validated.csv'
    )
    uc = ursgal.UController(
        params=params
    )

    filtered_csv = uc.execute_misc_engine(
        input_file=csv_file_to_filter,
        engine='filter_csv_1_0_0',
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the unify_csv function for omssa engine

'''
import ursgal
import csv
import pickle
import os

modifications = [
    'M,opt,any,Oxidation',        # Met oxidation
    'C,fix,any,Carbamidomethyl',  # Carbamidomethylation
    '*,opt,Prot-N-term,Acetyl',    # N-Acteylation
    # '*,opt,any,Deamidated:18O(1)'  # test case for ':' in unimod, is Deamidated:18O(1), with heavy oxygen, but this fails right now!!!
    'K,opt,any,Label:13C(5)15N(1)', # we have to test SILAC!!!!
    'K,opt,any,Label:13C(6)15N(2)' # we have to test SILAC!!!!
]

R = ursgal.UController(
    profile = 'LTQ XL low res',
    params  = {
        'database': os.path.join(
            'tests',
            'data',
            'BSA.fasta'
        ),
        'modifications' : modifications
    },
    force  = False
)
R.map_mods()

scan_rt_lookup = pickle.load(
    open(
        os.path.join(
            'tests',
            'data',
            '_test_ursgal_lookup.pkl'
        )
        ,
        'rb'
    )
)

unify_csv_main = R.unodes['unify_csv_1_0_0']['class'].import_engine_as_python_function()

input_csv = os.path.join(
    'tests',
    'data',
    'omssa_2_1_9',
    'test_BSA1_omssa_2_1_9.csv'
)
output_csv = os.path.join(
    'tests',
    'data',
    'omssa_2_1_9',
    'test_BSA1_omssa_2_1_9_unified.csv'
)

unify_csv_main(
    input_file     = input_csv,
    output_file    = output_csv,
    scan_rt_lookup = scan_rt_lookup,
    params = {
        'translations' : {
            'decoy_tag': 'decoy_',
            'enzyme' : 'KR;C;P',
            'semi_enzyme' : False,
            'database': os.path.join( 'tests', 'data', 'BSA.fasta'),
            'protein_delimiter' : '<|>',
            'psm_merge_delimiter' : ';',
            'keep_asp_pro_broken_peps':True,
            'precursor_mass_tolerance_minus': 5,
            'precursor_mass_tolerance_plus' : 5,
            'precursor_isotope_range' : "0,1",
            'max_missed_cleavages' : 2,
            'rounded_mass_decimals' : 3,
            'use_pyqms_for_mz_calculation': False,
            'aa_exception_dict' : {
                'J' : {
                    'original_aa' : ['L','I'],
                },
                'O' : {
                    'original_aa' : ['K'],
                    'unimod_name' : 'Methylpyrroline',
                },
            },
        },
        'label' : '',
        'mods' : R.params['mods'],
    },
    search_engine  = 'omssa_2_1_9',
)

ident_list = [ ]
for line_dict in csv.DictReader(open(output_csv, 'r')):
    ident_list.append( line_dict )
# print(ident_list)

def unify_omssa_test():
    for test_id, test_dict in enumerate(ident_list):
        yield unify_omssa, test_dict


def unify_omssa( test_dict ):
    assert 'uCalc m/z' in test_dict.keys()

    for key in [
            'Spectrum ID',
            'Spectrum Title',
            'Retention Time (s)',
            'Modifications',
            'uCalc m/z',
            'Sequence',
            'Is decoy',
            'Protein ID',
            'Sequence Start',
            'Sequence Stop',
            'Sequence Pre AA',
            'Sequence Post AA',
            'Complies search criteria'
        ]:
        test_value = test_dict[key]
        expected_value = test_dict['Expected {0}'.format(key)]
        if key in ('Retention Time (s)', 'uCalc m/z'):
            test_value      = round(float(test_value), 4)
            expected_value  = round(float(expected_value), 4)

        assert test_value == expected_value, '''
  Unexpected value in column "{0}":
    test value:     {1}
    expected value: {2}
            '''.format(key, test_value, expected_value)


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(ident_list):
        unify_omssa(test_dict)



#!/usr/bin/env python
from .msamanda_1_0_0_5243 import msamanda_1_0_0_5243 as msamanda


class msamanda_1_0_0_7503( msamanda ):
    """
    MSAmanda 1_0_0_7503 UNode

    Import functions from msamanda_1_0_0_5243
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'MSAmanda',
        'version'            : '1.0.0.7503',
        'release_date'       : None,
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'input_extensions'   : ['.mgf'],
        'output_extensions'  : ['.csv'],
        'create_own_folder'  : True,
        'include_in_git'     : None,
        'distributable'  : False,
        'in_development'     : False,
        'utranslation_style' : 'msamanda_style_1',
        'engine' : {
            'win32' : {
                '64bit' : {
                    'exe'            : 'MSAmanda.exe',
                    'url'            : '',
                    'zip_md5'        : None,
                    'additional_exe' : [],
                },
            },
        },
        'citation' : \
            'Dorfer V, Pichler P, Stranzl T, Stadlmann J, Taus T, Winkler S, '\
            'Mechtler K. (2014) MS Amanda, a universal identification '\
            'algorithm optimized for high accuracy tandem mass spectra.',
    }
    pass



#!/usr/bin/env python
import ursgal
import os
import pprint
from collections import defaultdict as ddict
import csv
import shutil

class quameter_1_1_10165( ursgal.UNode ):
    """
    QuaMeter unode

    Note:
        Please download and install Quameter manually. It comes with Proteowizard
        Bumbershoot.

    Warning:
        Still in testing phase

    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'QuaMeter',
        'version'                     : '1.1.10165',
        'release_date'                : '2016-11-07',
        'utranslation_style'          : 'quameter_style_1',
        'input_extensions'            : ['.raw', '.mzML'],
        'output_extensions'           : ['.tsv'],
        'create_own_folder'           : False,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'           : False,
        'engine_type'                 : {
            'misc_engine' : True,
        },
        'engine'                      : {
            'linux'    : {
                '64bit' : {
                    'exe'            : 'quameter',
                    'url'            : '',
                    'zip_md5'        : '',
                    'additional_exe' : [],
                },
            },
        },
        'citation'                   : \
            'Ma, Z. Q., Polzin, K. O., Dasari, S., Chambers, M. C., Schilling, B., '
            'Gibson, B. W., Tran, B. Q., Vega-Montoto, L., Liebler, D. C., and Tabb, D. L. '
            '(2012) QuaMeter: Multivendor performance metrics for LC-MS/MS proteomics '
            'instrumentation. Anal. Chem. 84, 5845–5850'
    }

    def __init__(self, *args, **kwargs):
        super(quameter_1_1_10165, self).__init__(*args, **kwargs)
        pass

    def write_params_file(self):

        return 

    def preflight( self ):
        '''
        Formatting the command line via self.params

        Returns:
                dict: self.params
        '''

        # pprint.pprint(self.params['translations']['_grouped_by_translated_key'])
        # pprint.pprint(self.params)
        # exit()
        
        self.input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )
        if self.input_file.lower().endswith('.mzml') or \
            self.input_file.lower().endswith('.raw'):
            self.params['translations']['mzml_input_file'] = self.input_file
        elif self.input_file.lower().endswith('.mgf'):
            self.params['translations']['mzml_input_file'] = \
                self.meta_unodes['ucontroller'].get_mzml_that_corresponds_to_mgf( self.input_file )
            self.print_info(
                'QuaMeter can only read RAW or mzML,'
                'thecorresponding mzML file {0} will be used instead.'.format(
                    os.path.abspath(self.params['translations']['mzml_input_file'])
                ),
                caller = "INFO" 
            )
        else:
            raise Exception('QuaMeter input spectrum file must be in mzML or RAW format!')

        self.params[ 'command_list' ] = [
            self.exe,
            self.params['translations']['mzml_input_file'],
            '-MetricsType',
            'idfree'
        ]

        self.params['translations']['output_file_incl_path'] = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        return self.params

    def postflight(self):
        '''
        Read tsvs and write final output file
        '''
        output_file_2_move = self.params['translations']['mzml_input_file'].replace(
            '.mzML',
            '.qual.tsv'
        )
        shutil.move(
            output_file_2_move,
            self.params['translations']['output_file_incl_path']
        )       
        return



#!/usr/bin/env python
import ursgal
import os
import pprint
from collections import defaultdict as ddict
import csv
import itertools
import sys

from .msfragger_20170103 import msfragger_20170103 as msfragger

class msfragger_20171106( msfragger ):
    """
    MSFragger unode

    Note:
        Please download and install MSFragger manually from
        http://www.nesvilab.org/software.html

    Reference:
    Kong, A. T., Leprevost, F. V, Avtonomov, D. M., Mellacheruvu, D., and Nesvizhskii, A. I. (2017)
    MSFragger: ultrafast and comprehensive peptide identification in mass spectrometry–based
    proteomics. Nature Methods 14

    Note:
        Addition of user amino acids not implemented yet. Only mzML search
        possible at the moment. The mgf file can still be passed to the node,
        but the mzML has to be in the same folder as the mgf.

    Warning:
        Still in testing phase!
        Metabolic labeling based 15N search may still be errorprone. Use with
        care!

    """
    META_INFO = {
        'edit_version'                : 1.00,
        'name'                        : 'MSFragger',
        'version'                     : '20171106',
        'release_date'                : '2017-11-06',
        'utranslation_style'          : 'msfragger_style_1',
        'input_extensions'            : ['.mgf', '.mzML', '.mzXML'],
        'output_extensions'           : ['.csv'],
        'create_own_folder'           : True,
        'in_development'              : False,
        'include_in_git'              : False,
        'distributable'           : False,
        'engine_type' : {
            'protein_database_search_engine' : True,
        },
        'engine'                      : {
            'platform_independent'    : {
                'arc_independent' : {
                    'exe'            : 'MSFragger-20171106.jar',
                    'url'            : 'http://www.nesvilab.org/software.html',
                    'zip_md5'        : '',
                    'additional_exe' : [],
                },
            },
        },
        'citation'                   : \
            'Kong, A. T., Leprevost, F. V, Avtonomov, '
            'D. M., Mellacheruvu, D., and Nesvizhskii, A. I. (2017) MSFragger: '
            'ultrafast and comprehensive peptide identification in mass '
            'spectrometry-based proteomics. Nature Methods 14'
    }


"""Read mgf file and write RT lookup pickle
"""
import hashlib
import os
import pickle
import re


def md5(fname):
    # from https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def generate_spectra(fh):
    """Generate spectra from openend mgf handle.

    Args:
        fh (IO): file obj
    Yields:
        spec_dict (dict): Dict contaning spec data
    """
    for line in fh:
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith('BEGIN IONS'):
            spec_dict = {
                'data': []
            }
        elif '=' in line:
            key, val = line.split('=', 1)
            spec_dict[key] = val
        elif line[0].isdigit():
            mz, i = line.strip().split()
            spec_dict['data'].append(
                (float(mz), float(i))
            )
        elif line.startswith('END IONS'):
            yield spec_dict


def main(input_file, output_file, rt_pickle_name, spec_id_regex, run_id_regex):
    """ead mgf file and write RT lookup pickle.

    Args:
        input_file (str): MGF input file path
        output_file (str): Outputfile name
        rt_pickle_name (str): Name of the RT pickle file
        spec_id_regex (str): Regex to extract spec ID from mgf title
        run_id_regex (str): Regex to extract run ID from mgf title

    Returns:
        Outputfile with RT pickle md5 hash
    """
    PRINT_MISSING_RT_WARNING = True
    lookup_path = os.path.join(
        os.path.dirname(input_file),
        rt_pickle_name
    )
    if os.path.exists(lookup_path):
        with open(lookup_path, 'rb') as fin:
            rt_dict = pickle.load(fin)
    else:
        rt_dict = {}
    input_base_wo_suffix, suff = os.path.splitext(os.path.basename(input_file))
    if input_base_wo_suffix not in rt_dict:
        tmp = {
            'rt_2_scan': {},
            'scan_2_rt': {},
            'scan_2_mz': {},
            'unit': 'minute'  # in mgf, its always seconds, but to stay compatible ..
        }
        spec_id_regex = re.compile(spec_id_regex)

        # rt = 0   # dev mgf has no rt ...

        with open(input_file) as fin:
            for spec_dict in generate_spectra(fin):
                spec_id = spec_id_regex.match(spec_dict['TITLE']).group(1)
                rt = float(spec_dict['RTINSECONDS']) / 60
                precursor_mz = float(spec_dict['PEPMASS'].strip())

                tmp['rt_2_scan'][rt] = int(spec_id)
                tmp['scan_2_rt'][int(spec_id)] = rt
                tmp['scan_2_mz'][int(spec_id)] = precursor_mz
        rt_dict[input_base_wo_suffix] = tmp
        with open(lookup_path, 'wb') as fout:
            pickle.dump(rt_dict, fout)
    with open(output_file, 'wt') as fout:
        md5_sum = md5(lookup_path)
        fout.write(md5_sum)
    return output_file



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os
import sys


def main():
    '''
    Simple example script how to generate a target decoy database using one or
    multiple fasta files as input.

    Note:
        A 'shuffled peptide preserving cleavage sites' database is 
        generated.

    usage:

        ./generate_target_decoy_from_fasta.py <fasta_database_1> <fasta_database_2> <fasta_database_n> ...

    '''
    params = {
        'enzyme': 'trypsin',
        # 'decoy_generation_mode': 'reverse_protein',
    }

    fasta_database_list = sys.argv[1:-1]
    target_db_name = sys.argv[-1]

    uc = ursgal.UController(
        params = params
    )

    new_target_decoy_db_name = uc.execute_misc_engine(
        input_file      = fasta_database_list,
        engine           = 'generate_target_decoy_1_0_0',
        output_file_name = target_db_name,
    )
    print(
        'Generated target decoy database: {0}'.format(
            new_target_decoy_db_name
        )
    )

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print(main.__doc__)
    main()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Executes a search with a current maximum of 3 versions of MSGF+ on an
    example file from the data from Barth et al. (2014)

    usage:
        ./msgf_version_comparison.py


    This is a simple example file to show the straightforward comparison of
    different program versions of MSGF+

    Creates a Venn diagram with the peptides obtained by the different
    versions.

    An example plot can be found in the online documentation.

    Note:
        Uses the new MS-GF+ C# mzid converter if available

    '''

    params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ],
        'ftp_url': 'ftp.peptideatlas.org',
        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',
        'ftp_include_ext': [
            'JB_FASP_pH8_2-3_28122012.mzML',
        ],
        'ftp_output_folder': os.path.join(
            os.pardir,
            'example_data',
            'msgf_version_comparison'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        ),
        'remove_temporary_files': False,

    }

    if os.path.exists(params['ftp_output_folder']) is False:
        os.mkdir(params['ftp_output_folder'])

    uc = ursgal.UController(
        profile='LTQ XL low res',
        params=params
    )

    engine_list = [
        'msgfplus_v9979',
        'msgfplus_v2016_09_16',
    ]
    if 'msgfplus_v2017_01_27' in uc.unodes.keys():
        if uc.unodes['msgfplus_v2017_01_27']['available']:
            engine_list.append('msgfplus_v2017_01_27')
    if 'msgfplus_v2018_01_30' in uc.unodes.keys():
        if uc.unodes['msgfplus_v2018_01_30']['available']:
            engine_list.append('msgfplus_v2018_01_30')

    if 'msgfplus_C_mzid2csv_v2017_07_04' in uc.unodes.keys():
        if uc.unodes['msgfplus_C_mzid2csv_v2017_07_04']['available']:
            uc.params['msgfplus_mzid_converter_version'] = 'msgfplus_C_mzid2csv_v2017_07_04'

    mzML_file = os.path.join(
        params['ftp_output_folder'],
        params['ftp_include_ext'][0]
    )
    if os.path.exists(mzML_file) is False:
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    filtered_files_list = []
    for engine in engine_list:
        unified_result_file = uc.search(
            input_file=mzML_file,
            engine=engine,
        )

        validated_file = uc.validate(
            input_file=unified_result_file,
            engine='percolator_2_08',
        )

        filtered_file = uc.execute_misc_engine(
            input_file=validated_file,
            engine='filter_csv_1_0_0',
        )

        filtered_files_list.append(filtered_file)

    uc.visualize(
        input_files=filtered_files_list,
        engine='venndiagram_1_1_0',
    )
    return

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
R = ursgal.UController()

TESTS = [
    {
        'input' : 'test.json',
        'output' : (False, 'test.json')
    },
    {
        'input' : ['test.json'],
        'output' : (False, 'test.json')
    },
    {
        'input' : ['test.json', '2131'],
        'output' : (True, ['test.json', '2131'])
    },
    # {
    #     'input' : 'test_without_database.json',
    #     'output' : '0cdd52e86608855ac678e212bedd0743'
    # }
]


def distinguish_multi_and_single_input_test():
    for test_id, test_dict in enumerate(TESTS):
        yield distinguish_multi_and_single_input, test_dict


def distinguish_multi_and_single_input( test_dict ):
    out_put = R.distinguish_multi_and_single_input( test_dict['input'] )
    print( out_put , test_dict)
    assert out_put == test_dict['output'], 'MD5 {0} failed'.format(
        test_dict
    )


if __name__ == '__main__':
    for test_id, test_dict in enumerate(TESTS):
        distinguish_multi_and_single_input( test_dict )



#!/usr/bin/env python3
# encoding: utf-8
'''

Test the peptide_regex function

'''
import ursgal
import os

database = os.path.join( 'tests', 'data', 'BSA.fasta')

R = ursgal.UController(
    profile = 'LTQ XL low res',
    params  = {
        'database' : database
    },
    force   = False
)

TESTS = [
    {
        'peptide' : 'SHCIAEVEK',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : 310,
        'stop1'    : 318,
        'pre_aa1'  : 'K',
        'post_aa1' : 'D'
    },
    {
        'peptide' : 'SHCIAEVE',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4', # test short id
        'start1'   : 310,
        'stop1'    : 317,
        'pre_aa1'  : 'K',
        'post_aa1' : 'K'
    },
    {
        'peptide' : 'YOLO',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : None,
        'stop1'    : None,
        'pre_aa1'  : None,
        'post_aa1' : None
    },
    {
        'peptide' : 'SHXIAEVEK',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : None,
        'stop1'    : None,
        'pre_aa1'  : None,
        'post_aa1' : None
    },

    {
        'peptide' : 'EYEATLEECCAK',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : 375,
        'stop1'    : 386,
        'pre_aa1'  : 'K',
        'post_aa1' : 'D'
    },
    {
        'peptide' : 'LR',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : 104,
        'stop1'    : 105,
        'pre_aa1'  : 'S',
        'post_aa1' : 'E',
        'start2'   : 221,
        'stop2'    : 222,
        'pre_aa2'  : 'R',
        'post_aa2' : 'C',
        'start3'   : 370,
        'stop3'    : 371,
        'pre_aa3'  : 'L',
        'post_aa3' : 'L'
    },
    {
        'peptide' : 'MKWVTFISLLLLFSSAYSR',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : 1,
        'stop1'    : 19,
        'pre_aa1'  : '-',
        'post_aa1' : 'G'
    },
    {
        'peptide' : 'LVVSTQTALA',
        'protein_id1':'sp|P02769|ALBU_BOVIN Serum albumin OS=Bos taurus GN=ALB PE=1 SV=4',
        'start1'   : 598,
        'stop1'    : 607,
        'pre_aa1'  : 'K',
        'post_aa1' : '-'
    },

]

def peptide_regex_test():
    for test_id, test_dict in enumerate(TESTS):
        yield peptide_regex, test_dict

def peptide_regex( test_dict ):
    '''
    peptide_regex(self, database, protein_id, peptide):
    '''
    return_list = R.peptide_regex(
        database,
        test_dict['protein_id1'],
        test_dict['peptide'],
    )
    print(return_list)
    for n, proteins in enumerate(return_list):
        for m, pep_regex in enumerate(proteins): 
            start,stop,pre_aa,post_aa, protein_id = pep_regex
            print(test_dict)
            print(start,stop,pre_aa,post_aa,protein_id)
            print()
            assert start      == test_dict['start{0}'.format(m+1)]
            assert stop       == test_dict['stop{0}'.format(m+1)]
            assert pre_aa     == test_dict['pre_aa{0}'.format(m+1)]
            assert post_aa    == test_dict['post_aa{0}'.format(m+1)]
            assert protein_id == test_dict['protein_id{0}'.format(n+1)]


if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(TESTS):
        peptide_regex(test_dict)




#!/usr/bin/env python3
# encoding: utf-8
import ursgal
import csv
from collections import defaultdict as ddict
import os
import glob
import math

params = {
    'database': os.path.join(
        os.pardir,
        'example_data',
        'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
    ),
    'csv_filter_rules': [
        ['Is decoy', 'equals', 'false'],
        ['PEP', 'lte', 0.01],
    ],
}
# We specify all search engines and validation engines that we want
# to use in a list (version numbers might differ on windows or mac):
search_engines = [
    'omssa',
    'xtandem_piledriver',
    'msgfplus_v9979',
    # 'myrimatch_2_1_138',
    # 'msamanda_1_0_0_5243',
]

validation_engines = [
    'percolator_2_08',
    'qvality',
]

# The different levels with different modifications
# for the cascade are defined
cascade = {
    '0': [
        'C,fix,any,Carbamidomethyl'
    ],
    '1': [
        'C,fix,any,Carbamidomethyl',
        'N,opt,any,Deamidated',
        'Q,opt,any,Deamidated'
    ],
    '2': [
        'C,fix,any,Carbamidomethyl',
        '*,opt,Prot-N-term,Acetyl'
    ],
    '3': [
        'C,fix,any,Carbamidomethyl',
        'M,opt,any,Oxidation'
    ],
    '4': [
        'C,fix,any,Carbamidomethyl',
        'E,opt,any,Methyl',
        'K,opt,any,Methyl',
        'R,opt,any,Methyl'
    ],
    '5': [
        'C,fix,any,Carbamidomethyl',
        'S,opt,any,Phospho',
        'T,opt,any,Phospho'
    ],
}

mass_spectrometer = 'LTQ XL low res'

get_params = {
    'ftp_url': 'ftp.peptideatlas.org',
    'ftp_login': 'PASS00269',
    'ftp_password': 'FI4645a',
    'ftp_include_ext': [
        'JB_FASP_pH8_2-3_28122012.mzML',
        'JB_FASP_pH8_2-4_28122012.mzML',
        'JB_FASP_pH8_3-1_28122012.mzML',
        'JB_FASP_pH8_4-1_28122012.mzML',
    ],
    'ftp_output_folder': os.path.join(
        os.pardir,
        'example_data',
        'cascade_search'
    ),
    'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
    'http_output_folder': os.path.join(
        os.pardir,
        'example_data'
    )
}


def get_files():
    uc = ursgal.UController(
        params=get_params
    )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    if os.path.exists(get_params['ftp_output_folder']) is False:
        os.makedirs(get_params['ftp_output_folder'])
    uc.fetch_file(
        engine='get_ftp_files_1_0_0'
    )

    spec_files = []
    for mzML_file in glob.glob(
        os.path.join(
            get_params['ftp_output_folder'],
            '*.mzML'
        )
    ):
        spec_files.append(mzML_file)

    return spec_files


def search(validation_engine):
    '''
    Executes a cascade search on four example files from the 
    data from Barth et al.

    usage:
        ./cascade_search_example.py

    Searches for peptides using a cascade search approach similar to Kertesz-Farkas et al.
    for which spectra were first searched for unmodified peptides, followed by consecutive searches
    for the following modifications: 
    oxidation of M,
    deamidation of N/Q,
    methylation of E/K/R,
    N-terminal acetylation,
    phosphorylation of S/T.
    After each step, spectra with a PSM below 1 % PEP were removed. 
    '''
    # Initializing the uPLANIT UController class with
    # our specified modifications and mass spectrometer
    uc = ursgal.UController(
        profile=mass_spectrometer,  # 'LTQ XL low res' profile!
        params=params
    )
    # complete workflow for every level of the cascade:
    # every spectrum file is searched with every search engine,
    # results are validated seperately,
    # validated results are merged and filtered for targets and PEP <= 0.01.

    def workflow(spec_file, prefix=None, validation_engine=None, filter_before_validation=False, force=False):
        validated_results = []

        # Convert mzML to MGF outside the loop, so this step is not repeated in
        # the loop
        mgf_spec_file = uc.convert(
            input_file=spec_file,
            engine='mzml2mgf_1_0_0'
        )
        for search_engine in search_engines:
            uc.params['prefix'] = prefix
            unified_search_results = uc.search(
                input_file=mgf_spec_file,
                engine=search_engine,
                force=force,
            )
            uc.params['prefix'] = ''

            if filter_before_validation == True:
                uc.params['csv_filter_rules'] = [
                    ['Modifications', 'contains', '{0}'.format(
                        cascade[level][1].split(',')[3])]
                ]
                filtered_search_results = uc.execute_misc_engine(
                    input_file=unified_search_results,
                    engine='filter_csv_1_0_0'
                )
            else:
                filtered_search_results = unified_search_results
            validated_search_results = uc.validate(
                input_file=filtered_search_results,
                engine=validation_engine,
                force=force,
            )
            validated_results.append(validated_search_results)

        validated_results_from_all_engines = uc.execute_misc_engine(
            input_file=sorted(validated_results),
            engine='merge_csvs_1_0_0',
            force=force,
        )
        uc.params['csv_filter_rules'] = [
            ['Is decoy', 'equals', 'false'],
            ['PEP', 'lte', 0.01],
        ]
        filtered_validated_results = uc.execute_misc_engine(
            input_file=validated_results_from_all_engines,
            engine='filter_csv_1_0_0'
        )
        return filtered_validated_results

    result_files = []
    for spec_file in spec_files:
        spectra_with_PSM = set()
        for level in sorted(cascade.keys()):
            uc.params['modifications'] = cascade[level]
            if level == '0':
                results = workflow(
                    spec_file,
                    validation_engine=validation_engine,
                    prefix='cascade-lvl-{0}'.format(level)
                )
            else:
                uc.params['scan_exclusion_list'] = list(spectra_with_PSM)
                results = workflow(
                    spec_file,
                    validation_engine=validation_engine,
                    filter_before_validation=True,
                    force=True,
                    prefix='cascade-lvl-{0}'.format(level)
                )
            result_files.append(results)
            #  spectrum IDs for PSMs are written into an exclusion list for the next level of the cascade search,
            #  these spectra will b excluded during mzml2mgf conversion
            with open(results) as in_file:
                csv_input = csv.DictReader(in_file)
                for line_dict in csv_input:
                    spectra_with_PSM.add(line_dict['Spectrum ID'])
            print(
                'Number of spectra that will be removed for the next cacade level: {0}'.format(
                    len(spectra_with_PSM)
                )
            )

    results_all_files = uc.execute_misc_engine(
        input_file=sorted(result_files),
        engine='merge_csvs_1_0_0',
    )
    return results_all_files


def analyze(collector):
    '''
    Simle analysis script for the cascade search,
    counting the number of identified peptides (combination of peptide sequence and modifications)
    and PSMs (additionally include the spectrum ID)
    '''

    mod_list = [
        'Oxidation',
        'Deamidated',
        'Methyl',
        'Acetyl',
        'Phospho'
    ]
    fieldnames = [
        'approach',
        'count_type',
        'validation_engine',
        'unmodified',
        'multimodified'
    ] + mod_list + ['total']

    csv_writer = csv.DictWriter(open('cascade_results.csv', 'w'), fieldnames)
    csv_writer.writeheader()
    uc = ursgal.UController()
    uc.params['validation_score_field'] = 'PEP'
    uc.params['bigger_scores_better'] = False

    # Count the number of identified peptides and PSMs for the different modifications
    # Spectra with multiple PSMs are sanitized, i.e. only the PSM with best PEP score is counted
    # and only if the best hit has a PEP that is at least two orders of
    # magnitude smaller than the others
    for validation_engine, result_file in collector.items():
        counter_dict = {
            'psm': ddict(set),
            'pep': ddict(set)
        }
        grouped_psms = uc._group_psms(
            result_file, validation_score_field='PEP', bigger_scores_better=False)
        for spec_title, grouped_psm_list in grouped_psms.items():
            best_score, best_line_dict = grouped_psm_list[0]
            if len(grouped_psm_list) > 1:
                second_best_score, second_best_line_dict = grouped_psm_list[1]
                best_peptide_and_mod = best_line_dict[
                    'Sequence'] + best_line_dict['Modifications']
                second_best_peptide_and_mod = second_best_line_dict[
                    'Sequence'] + second_best_line_dict['Modifications']

                if best_peptide_and_mod == second_best_peptide_and_mod:
                    line_dict = best_line_dict
                elif best_line_dict['Sequence'] == second_best_line_dict['Sequence']:
                    if best_score == second_best_score:
                        line_dict = best_line_dict
                    else:
                        if (-1 * math.log10(best_score)) - (-1 * math.log10(second_best_score)) >= 2:
                            line_dict = best_line_dict
                        else:
                            continue
                else:
                    if (-1 * math.log10(best_score)) - (-1 * math.log10(second_best_score)) >= 2:
                        line_dict = best_line_dict
                    else:
                        continue
            else:
                line_dict = best_line_dict

            count = 0
            for mod in mod_list:
                if mod in line_dict['Modifications']:
                    count += 1
            key_2_add = ''
            if count == 0:
                key_2_add = 'unmodified'
            elif count >= 2:
                key_2_add = 'multimodified'
            elif count == 1:
                for mod in mod_list:
                    if mod in line_dict['Modifications']:
                        key_2_add = mod
                        break
            # for peptide identification comparison
            counter_dict['pep'][key_2_add].add(
                line_dict['Sequence'] + line_dict['Modifications']
            )
            # for PSM comparison
            counter_dict['psm'][key_2_add].add(
                line_dict['Spectrum Title'] +
                line_dict['Sequence'] + line_dict['Modifications']
            )
        for counter_key, count_dict in counter_dict.items():
            dict_2_write = {
                'approach': 'cascade',
                'count_type': counter_key,
                'validation_engine': validation_engine
            }
            total_number = 0
            for key, obj_set in count_dict.items():
                dict_2_write[key] = len(obj_set)
                total_number += len(obj_set)
            dict_2_write['total'] = total_number
            csv_writer.writerow(dict_2_write)
    return

if __name__ == '__main__':
    spec_files = get_files()
    collector = {}
    for validation_engine in validation_engines:
        results_all_files = search(validation_engine)
        print('>>> ', 'final results for {0}'.format(
            validation_engine), ' were written into:')
        print('>>> ', results_all_files)
        collector[validation_engine] = results_all_files
    analyze(collector)
    print('>>> ', 'number of identified peptides and PSMs were written into:')
    print('>>> ', 'cascade_results.csv')



#!/usr/bin/env python
'''
Filters the result csvs

usage:
    ./filter_csv_1_0_0.py <input_file> <output_file>


'''

from __future__ import print_function
import sys
import os
import csv
import ursgal
import re
import operator

if sys.platform != 'win32':
    csv.field_size_limit(sys.maxsize)


def main( input_file=None, output_file=None, filter_rules=None, output_file_unfiltered = None):
    '''
    Filters csvs
    '''
    if filter_rules is None:
        print('No filter rules defined! Exiting now...')
        sys.exit(1)
    elif len(filter_rules) == 0:
        print('No filter rules defined! Exiting now...')
        sys.exit(1)
    else:
        pass

    csv_kwargs = {}
    if sys.platform == 'win32':
        csv_kwargs['lineterminator'] = '\n'
    else:
        csv_kwargs['lineterminator'] = '\r\n'

    output_file_object = open(output_file,'w')

    if output_file_unfiltered is not None:
        unfiltered_output_file_object = open(
            output_file_unfiltered,
            'w'
        )

    with open( input_file, 'r' ) as in_file:
        csv_input  = csv.DictReader( in_file )
        csv_output = csv.DictWriter(
            output_file_object,
            list(csv_input.fieldnames),
            **csv_kwargs
        )
        csv_output.writeheader()
        if output_file_unfiltered is not None:
            unfiltered_csv_output = csv.DictWriter(
                unfiltered_output_file_object,
                list(csv_input.fieldnames),
                **csv_kwargs
            )
            unfiltered_csv_output.writeheader()
        for line_pos, line_dict in enumerate(csv_input):
            write_row_bools = set()
            for rule_tuple in filter_rules:
                dict_key, rule, value = rule_tuple
                if dict_key not in line_dict.keys():
                    print(
                        '''Rule not recognized: {0}
Input file: {1}
headers : {2}
Cause: Specified key not in csv fieldnames
Did you misspell the field name?'''.format(
                            rule_tuple,
                            input_file,
                            csv_input.fieldnames,

                        )
                    )
                    raise Exception
                else:
                    if rule in ['lte', 'gte', 'lt' , 'gt']:
                        # requires not None! and to be floatable
                        if line_dict[ dict_key ] is None:
                            write_row_bools.add(False)

                        elif rule == 'lte':
                            cpm_method = operator.__le__
                        elif rule == 'gte':
                            cpm_method = operator.__ge__
                        elif rule == 'lt':
                            cpm_method = operator.__lt__
                        elif rule == 'gt':
                            cpm_method = operator.__gt__

                        try:
                            floated_value = float(line_dict[dict_key])
                        except:
                            if line_dict[ dict_key ] is '':
                                floated_value = None
                            else:
                                print('[ ERROR ] Value to be filtered could not be converted to float')
                                print('Value:'+line_dict[dict_key])
                                sys.exit(1)
                                
                        if floated_value is None:
                            write_row_bools.add(False)
                        elif cpm_method( floated_value, value):
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'equals':
                        if value == line_dict[dict_key]:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'equals_not':
                        if value != line_dict[dict_key]:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'contains':
                        if value in line_dict[dict_key]:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'contains_not':
                        if line_dict[dict_key] is None:
                            write_row_bools.add(True)
                        elif value not in line_dict[dict_key]:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'regex':
                        if re.search(value, line_dict[dict_key]) is not None:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'contains_glycosite':
                        if re.search(value, line_dict[dict_key]) is not None:
                            write_row_bools.add(True)
                        elif line_dict[dict_key][-2] == 'N' and line_dict[dict_key][-1] != 'P':
                            if 'S' in line_dict['Sequence Post AA'] or 'T' in line_dict['Sequence Post AA']:
                                write_row_bools.add(True)
                            else:
                                write_row_bools.add(False)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'mod_at_glycosite':
                        mods =  line_dict[dict_key].split(';')
                        accepted = False
                        for mod in mods:
                            if value in mod:
                                pos = int(mod.split(':')[-1])
                                if line_dict['Sequence'][pos-1] != 'N':
                                    continue
                                if line_dict['Sequence'][pos] == 'P':
                                    continue
                                if pos >= len(line_dict['Sequence'])-2:
                                    if 'S' not in line_dict['Sequence Post AA'] and 'T' not in line_dict['Sequence Post AA']:
                                        continue
                                else:
                                    if line_dict['Sequence'][pos+1] not in ['S', 'T']:
                                        continue
                                accepted = True
                            else:
                                continue
                        if accepted is True:
                            write_row_bools.add(True)
                        else:
                            write_row_bools.add(False)

                    elif rule == 'contains_element_of_list':
                        assert type(value) == list, '''
                        The value for the filter rule 'contains_element_of_list'
                        needs to be a list. You specified:
                        {0}
                        '''.format(value)
                        write = False
                        for element in value:
                            if element in line_dict[dict_key]:
                                write = True
                        write_row_bools.add(write)

                    else:
                        print('Rule: {0} not defined'.format(rule))
                        raise Exception

            if len(write_row_bools) == 1 and list(write_row_bools)[0] == True:
                csv_output.writerow( line_dict )
            elif output_file_unfiltered is not None:
                unfiltered_csv_output.writerow( line_dict )
            # break

    output_file_object.close()
    in_file.close()
    return output_file


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print(__doc__)
        sys.exit(1)

    if sys.argv[3] == 'None':
        output_file_unfiltered = None
    else:
        output_file_unfiltered = sys.argv[3]


    main(
        input_file     = sys.argv[1],
        output_file    = sys.argv[2],
        filter_rules   =[
            # ('PEP','lte',0.01),
            ('Is decoy','equals_not','true'),
            # ('Sequence','contains','T'),
            # ('Sequence','contains','S'),
            # ('Sequence','contains','R'),

        ],
        output_file_unfiltered = output_file_unfiltered
    )



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys

class mzml2mgf_1_0_0( ursgal.UNode ):
    """
    mzml2mgf_1_0_0 UNode

    Converts .mzML files into .mgf files
    """
    META_INFO = {
        'edit_version'       : 1.00,
        'name'               : 'mzml2mgf',
        'version'            : '1.0.0',
        'release_date'       : None,
        'engine_type' : {
            'converter' : True,
        },
        'input_extensions'   : ['.mzML', '.mzML.gz'],
        'output_extensions'  : ['.mgf'],
        'output_suffix'      : None,
        'in_development'     : False,
        'include_in_git'     : True,
        'distributable'      : True,
        'utranslation_style' : 'mzml2mgf_style_1',
        'engine' : {
            'platform_independent' : {
                'arc_independent' : {
                    'exe' : 'mzml2mgf_1_0_0.py',
                },
            },
        },
        'citation' : \
            'Kremer, L. P. M., Leufken, J., Oyunchimeg, P., Schulze, S. & '\
            'Fufezan, C. (2016) Ursgal, Universal Python Module Combining '\
            'Common Bottom-Up Proteomics Tools for Large-Scale Analysis. J. '\
            'Proteome res. 15, 788-794.',
    }

    def __init__(self, *args, **kwargs):
        super(mzml2mgf_1_0_0, self).__init__(*args, **kwargs)

    def _execute( self ):
        print('[ -ENGINE- ] Executing conversion ..')
        # self.time_point(tag = 'execution')
        mzml2mgf_main = self.import_engine_as_python_function()


        # try:
        if True:
            tmp = mzml2mgf_main(
                mzml = os.path.join(
                    self.io['input']['finfo']['dir'],
                    self.io['input']['finfo']['file']
                ),
                mgf = os.path.join(
                    self.io['output']['finfo']['dir'],
                    self.io['output']['finfo']['file']
                ),
                i_decimals            = self.params['translations']['num_i_decimals'],
                mz_decimals           = self.params['translations']['num_mz_decimals'],
                machine_offset_in_ppm = self.params['translations']['machine_offset_in_ppm'],
                scan_exclusion_list   = self.params['translations']['scan_exclusion_list'],
                scan_inclusion_list   = self.params['translations']['scan_inclusion_list'],
                prefix                = self.params.get('prefix',None),
                scan_skip_modulo_step = self.params['translations']['scan_skip_modulo_step'],
                ms_level              = self.params['translations']['ms_level'],
            )
        # except KeyError:
        #     print('''

        #         OBO Version changed ? converter uses obo names instead of tags.
        #         This need to be updated in time ...

        #         ''')
        # self.print_execution_time(tag='execution')
        return tmp



#!/usr/bin/env python3
# encoding: utf-8

import ursgal
import os


def main():
    '''
    Script to compare Percolator versions 2.08 and 3.2 using MSFragger as
    search engine
    '''

    engine_list = [
        'xtandem_vengeance',
    ]

    params = {
        'database': os.path.join(
            os.pardir,
            'example_data',
            'Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta'
        ),
        'modifications': [],
        'csv_filter_rules': [
            ['PEP', 'lte', 0.01],
            ['Is decoy', 'equals', 'false']
        ],
        'ftp_url': 'ftp.peptideatlas.org',
        'ftp_login': 'PASS00269',
        'ftp_password': 'FI4645a',
        'ftp_include_ext': [
            'JB_FASP_pH8_2-3_28122012.mzML',
        ],
        'ftp_output_folder': os.path.join(
            os.pardir,
            'example_data',
            'percolator_version_comparison'
        ),
        'http_url': 'https://www.sas.upenn.edu/~sschulze/Creinhardtii_281_v5_5_CP_MT_with_contaminants_target_decoy.fasta',
        'http_output_folder': os.path.join(
            os.pardir,
            'example_data'
        ),
        'infer_proteins': False,
        'percolator_post_processing': 'tdc',
    }

    if os.path.exists(params['ftp_output_folder']) is False:
        os.mkdir(params['ftp_output_folder'])

    uc = ursgal.UController(
        profile='LTQ XL low res',
        params=params
    )
    mzML_file = os.path.join(
        params['ftp_output_folder'],
        params['ftp_include_ext'][0]
    )
    if os.path.exists(mzML_file) is False:
        uc.fetch_file(
            engine='get_ftp_files_1_0_0'
        )
    if os.path.exists(params['database']) is False:
        uc.fetch_file(
            engine='get_http_files_1_0_0'
        )

    validation_engine_list = [
        'percolator_2_08',
        'percolator_3_2_1'
    ]
    uc.params['visualization_label_positions'] = {}
    for n, vce in enumerate(validation_engine_list):
        uc.params['visualization_label_positions'][str(n)] = vce
    for engine in engine_list:
        unified_search_results = uc.search(
            input_file=mzML_file,
            engine=engine,
        )

        validated_and_filtered_files_list = []
        for pc_version in validation_engine_list:
            uc.params['prefix'] = pc_version
            validated_file = uc.validate(
                input_file=unified_search_results,
                engine=pc_version,
            )
            filtered_file = uc.execute_misc_engine(
                input_file=validated_file,
                engine='filter_csv'
            )
            validated_and_filtered_files_list.append(
                filtered_file
            )
        uc.visualize(
            input_files=validated_and_filtered_files_list,
            engine='venndiagram_1_1_0',
        )

    return

if __name__ == '__main__':
    main()
